/*

  $Id: Auth4cAD.cpp 689 2018-11-28 17:54:35Z john $

  Copyright (c) 2008, The 4c Company. All rights reserved.
*/

#include <Auth4cAD/Auth4cAD.h>

#include <Auth4cDal/Auth4cDB.h>

#include <Auth4cData/Team.h>
#include <Auth4cData/TeamMember.h>
#include <Auth4cData/Factory.h>
#include <Auth4cData/InitMap.h>
#include <Auth4cData/AuthenticationToken.h>
#include <Auth4cData/SessionString.h>

#include <Auth4cDal/TeamMapQuery.h>

#include <Base4c/Util.h>
#include <Base4c/EventLogger.h>

#include <librfcdal/SqlInteger.h>

#include <time.h>
#include <sstream>

using namespace Auth4cAD;

//////////////////////////////////////////////////////////////////////////
//
// Authenticator
//

Authenticator::Authenticator( const Auth4cData::InitMap& initAuthenticator,
								const Auth4cData::Domain& domain,
								const std::string& strDescription,
								int nAuthenticationLibrary )
	: Auth4cData::Authenticator( initAuthenticator,
									domain,
									strDescription,
									nAuthenticationLibrary )
{
	Auth4cData::InitMap::const_iterator iterInit;

#if defined (LINUX)
	iterInit = authenticatorInit().find( "LDAP" );
	if( authenticatorInit().end() != iterInit )
	{
		_strHost = iterInit->second;
	}
#elif defined (_WINDOWS)
	iterInit = authenticatorInit().find( "Host" );
	if( authenticatorInit().end() != iterInit )
	{
		_strHost = iterInit->second;
	}
#endif /* LINUX */
	iterInit = authenticatorInit().find( "BaseDN" );
	if( authenticatorInit().end() != iterInit )
	{
		_strBaseDN = iterInit->second;
	}

	iterInit = authenticatorInit().find( "User" );
	if( authenticatorInit().end() != iterInit )
	{
		_strSearchUser = iterInit->second;
	}

	iterInit = authenticatorInit().find( "Password" );
	if( authenticatorInit().end() != iterInit )
	{
		_strSearchPassword = iterInit->second;
	}

	iterInit = authenticatorInit().find( "Template" );
	if( authenticatorInit().end() != iterInit )
	{
		_strSearchTemplate = iterInit->second;
	}
}

Authenticator::Authenticator( const Authenticator& authenticator )
	: Auth4cData::Authenticator( authenticator ),
	_strHost( authenticator._strHost ),
	_strBaseDN( authenticator._strBaseDN ),
	_strSearchUser( authenticator._strSearchUser ),
	_strSearchPassword( authenticator._strSearchPassword ),
	_strSearchTemplate( authenticator._strSearchTemplate )
{
}

Authenticator::~Authenticator()
{
}

const Authenticator& Authenticator::operator=( const Authenticator& authenticator )
{
	if( this != &authenticator )
	{
		Auth4cData::Authenticator::operator=( authenticator );

		_strHost = authenticator._strHost;
		_strBaseDN = authenticator._strBaseDN;
		_strSearchUser = authenticator._strSearchUser;
		_strSearchPassword = authenticator._strSearchPassword;
		_strSearchTemplate = authenticator._strSearchTemplate;
	}

	return *this;
}

Auth4cData::AuthenticatedUser Authenticator::authenticate( const Auth4cData::User& user, const Auth4cData::AuthenticationToken& authToken, const Auth4cData::Profile& profile )
{
	LDAP* pld = NULL;
	Auth4cData::AuthenticatedUser userA;

	try
	{
		pld = connectLDAP();

		userA = authenticate( user, authToken, profile, pld );

		disconnectLDAP( pld );
	}
	catch( ... )
	{
		if( NULL != pld )
		{
			disconnectLDAP( pld );
		}

		throw;
	}

	return userA;
}

void Authenticator::user( Auth4cData::User& userF, bool bRecurse /*= true*/ )
{
	LDAP* pld = NULL;

	try
	{
		pld = connectLDAP();

		user( userF, pld );

		disconnectLDAP( pld );
	}
	catch( ... )
	{
		if( NULL != pld )
		{
			disconnectLDAP( pld );
		}

		throw;
	}
}

void Authenticator::team( Auth4cData::Team& teamF )
{
	LDAP* pld = NULL;

/* Start
	Network conectivity issue sometimes make it impossible to fully populate a team. The code for
	populating the team object iterates through all the domains to determine whether the domain has
	any members for the specified team.

	Where a team is sited totally within one domain, the population of the team can fail purely
	because a domain which rightly contains no members is unreachable.

	This new code ensures that if the team has no entry in the TeamMap for the current domain (a
	fair indicator that the domain has no members) it will not attempt to connect to the back end.
*/
	bool bFound( false );
	const Auth4cData::XTeamPtrMap& xteamS( Auth4cData::Factory::teams( teamF ) );

	for( Auth4cData::XTeamPtrMap::const_iterator iterXTeams = xteamS.begin();
			iterXTeams != xteamS.end();
			++iterXTeams )
	{
		if( iterXTeams->second->domain() == domain().domain() )
		{
			bFound = true;
		}
	}

	if( !bFound )
	{
		return;
	}
/* End */

	try
	{
		pld = connectLDAP();

		team( teamF, pld );

		disconnectLDAP( pld );
	}
	catch( const Auth4cData::Exception& e )
	{
		if( NULL != pld )
		{
			disconnectLDAP( pld );
		}

		std::ostringstream os;
		os << "Auth4cAD::Authenticator::team(): " << domain().domain() << ": " << e;
		Base4c::EventLogger::theLogger()->logError( os.str(), Base4c::EventLogger::data );

		throw;
	}
	catch( ... )
	{
		if( NULL != pld )
		{
			disconnectLDAP( pld );
		}

		std::ostringstream os;
		os << "Auth4cAD::Authenticator::team(): " << domain().domain() << ": ...";
		Base4c::EventLogger::theLogger()->logError( os.str(), Base4c::EventLogger::data );

		throw;
	}
}

LDAP* Authenticator::connectLDAP()
{
	LDAP* pld = NULL;
	int nRC( 0 );

#if defined (LINUX)
	nRC = ldap_initialize( &pld, _strHost.c_str() );
#elif defined (_WINDOWS)
	pld = ldap_openA( (const PCHAR)_strHost.c_str(), LDAP_PORT );
#endif
	if( nRC )
	{
#if defined (LINUX)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATORINITFAILED, ldap_err2string( nRC ) );
#elif defined (_WINDOWS)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATORINITFAILED, ldap_err2stringA( nRC ) );
#endif
	}
	else if( NULL == pld )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATORINITFAILED, "NULL LDAP." );
	}

	int version = LDAP_VERSION3;
	if ( ldap_set_option( pld, LDAP_OPT_PROTOCOL_VERSION, &version ) < 0 )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATORINITFAILED, "Cannot select v3." );
	}

#if defined (LINUX)
	struct berval bvAuth;
	bvAuth.bv_len = _strSearchPassword.length();
	bvAuth.bv_val = (char*)_strSearchPassword.c_str();
	nRC = ldap_sasl_bind_s( pld,
							_strSearchUser.c_str(),
							LDAP_SASL_SIMPLE,
							&bvAuth,
							NULL,
							NULL,
							NULL );
#elif defined (_WINDOWS)
	nRC = ldap_simple_bind_sA( pld,
							  (const PCHAR)_strSearchUser.c_str(),
							  (const PCHAR)_strSearchPassword.c_str() );
#endif /* LINUX */

	if( nRC )
	{
#if defined (LINUX)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2string( nRC ) );
#elif defined (_WINDOWS)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2stringA( nRC ) );
#endif /* LINUX */
	}

	return pld;
}

void Authenticator::disconnectLDAP( LDAP* pld )
{
#if defined (LINUX)
	ldap_unbind_ext_s( pld, NULL, NULL );
#elif defined (_WINDOWS)
	ldap_unbind_s( pld );
#endif /* LINUX */
}

Auth4cData::AuthenticatedUser Authenticator::authenticate( const Auth4cData::User& userA, const Auth4cData::AuthenticationToken& authToken, const Auth4cData::Profile& profile, LDAP* pld )
{
	if( NULL == pld )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "LDAP is NULL." );
	}

	Base4c::VariableMap variableS;
	variableS["USERID"] = userA.userId();
	std::string strWho( Base4c::expand( _strSearchTemplate, variableS ) );

// Find the user
	LDAPMessage* presult( NULL );
	LDAPMessage* psubresult( NULL );
	std::string strDNUser;

#if defined (LINUX)
	int nRC = ldap_search_ext_s( pld,
									_strBaseDN.c_str(),
									LDAP_SCOPE_SUBTREE,
									strWho.c_str(),
									NULL,
									0,
									NULL,
									NULL,
									NULL,
									100,
									&presult );
#elif defined (_WINDOWS)
	int nRC = ldap_search_sA( pld, (const PCHAR)_strBaseDN.c_str(), LDAP_SCOPE_SUBTREE, (const PCHAR)strWho.c_str(), NULL, 0, &presult );
#endif /* LINUX */

	if( LDAP_SUCCESS != nRC  )
	{
		if( NULL != presult )
			ldap_msgfree( presult );

#if defined (LINUX)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2string( nRC ) );
#elif defined (_WINDOWS)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2stringA( nRC ) );
#endif /* LINUX */
	}

	if( NULL != presult )
	{
		if( 1 != ldap_count_entries( pld, presult ) )
		{
			ldap_msgfree( presult );

			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Invalid credentials." );
		}

		psubresult = ldap_first_entry( pld, presult );
		while( NULL != psubresult )
		{
#if defined (LINUX)
			strDNUser = ldap_get_dn( pld, psubresult );
#elif defined (_WINDOWS)
			strDNUser = ldap_get_dnA( pld, psubresult );
#endif /* LINUX */

			psubresult = ldap_next_entry( pld, psubresult );
		}

		ldap_msgfree( presult );
	}

	presult = NULL;

	try
	{
		authenticate( strDNUser, authToken, pld );
	}
	catch( const Auth4cData::Exception& /*e*/ )
	{
		throw;
	}

	Auth4cData::User userLocal( userA );
	user( userLocal, pld );

	std::ostringstream os;
	os << &userLocal << time( NULL );
	Auth4cData::SessionString sesStr( os.str() );
	Auth4cData::AuthenticatedUser userAuth( userLocal, sesStr, profile );
	return userAuth;
}

void Authenticator::user( Auth4cData::User& userF, LDAP* pld, bool bTeams /*= true*/ )
{
	user( userF, userF.userId(), pld );
}

void Authenticator::team( Auth4cData::Team& team, LDAP* pld )
{
	if( NULL == pld )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "LDAP is NULL." );
	}

	std::string strCN;
	std::string strCNLead;
	int nCount( 0 );

	const Auth4cData::XTeamPtrMap& xteamS( Auth4cData::Factory::teams( team ) );

	for( Auth4cData::XTeamPtrMap::const_iterator iterXTeams = xteamS.begin();
			iterXTeams != xteamS.end();
			++iterXTeams )
	{
		if( iterXTeams->second->domain() == domain().domain() )
		{
			strCN = iterXTeams->second->alienGroup();
			strCNLead = iterXTeams->second->alienGroupLeadGroup();
			++nCount;
		}
	}

	if( 1 != nCount )
	{
		return;
	}

	std::vector<std::string> strGroupMemberS;
	Auth4cData::UserMap userS;

	strGroupMemberS = groupMembers( strCN, pld );
	for( std::vector<std::string>::const_iterator iter = strGroupMemberS.begin();
			iter != strGroupMemberS.end();
			++iter )
	{
		try
		{
			Auth4cData::User userF;
			std::string strUserId( *iter );
			user( userF, strUserId, pld, false );
			userS[userF.key()] = userF;
		}
		catch( ... )
		{
			// eat the exception.
			// This is most likely to happen when a user has been deleted but is
			// still listed as a member of the group.
		}
	}

	if( strCNLead.length() )
	{
		strGroupMemberS = groupMembers( strCNLead, pld );
		for( std::vector<std::string>::const_iterator iter = strGroupMemberS.begin();
				iter != strGroupMemberS.end();
				++iter )
		{
			try
			{
				Auth4cData::User userF;
				std::string strUserId( *iter );
				user( userF, strUserId, pld, false );

				Auth4cData::UserMap::const_iterator iterUser = userS.find( userF.key() );
				if( iterUser != userS.end() )
				{
					Auth4cData::TeamMember member( iterUser->second, true );
					team.members()[member.key()] = member;
				}
			}
			catch( ... )
			{
				// eat the exception.
				// This is most likely to happen when a user has been deleted but is
				// still listed as a member of the group.
			}
		}
	}

	for( Auth4cData::UserMap::const_iterator iterUser = userS.begin();
			iterUser != userS.end();
			++iterUser )
	{
		if( team.members().find( iterUser->second.key() ) == team.members().end() )
		{
			Auth4cData::TeamMember member( iterUser->second, false );
			team.members()[member.key()] = member;
		}
	}
}

bool Authenticator::authenticate( const std::string& strDN, const Auth4cData::AuthenticationToken& authToken, LDAP* pld )
{
	// check for empty password field
	std::string strAuthenticationToken = authToken.str();
	if( "" == strAuthenticationToken )
	{
#if defined (LINUX)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Password is empty." );
#elif defined (_WINDOWS)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Password is empty." );
#endif /* LINUX */
	}

	struct berval bvAuth;
	bvAuth.bv_len = strAuthenticationToken.length();
	bvAuth.bv_val = (char*)strAuthenticationToken.c_str();
#if defined (LINUX)
	int n = ldap_sasl_bind_s( pld,
							  strDN.c_str(),
							  LDAP_SASL_SIMPLE,
							  &bvAuth,
							  NULL,
							  NULL,
							  NULL );
#elif defined (_WINDOWS)
	int n = ldap_simple_bind_sA( pld,
							  (const PCHAR)strDN.c_str(),
							  (const PCHAR)strAuthenticationToken.c_str() );
#endif /* LINUX */

	if( n )
	{
#if defined (LINUX)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2string( n ) );
#elif defined (_WINDOWS)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2stringA( n ) );
#endif /* LINUX */
	}

	return true;
}

void Authenticator::user( Auth4cData::User& user, const std::string& strSearch, LDAP* pld, bool bTeams /*= true*/ )
{
	if( NULL == pld )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "LDAP is NULL." );
	}

	std::string strWhat;
	std::string strBase;
	std::string strDNUser;
#if defined (_WINDOWS)
	ULONG scope;
#else
	int scope;
#endif /* _WINDOWS */

	if( strSearch.find( "=" ) == std::string::npos )
	{
		Base4c::VariableMap variableS;
		variableS["USERID"] = strSearch;
		strWhat = Base4c::expand( _strSearchTemplate, variableS );
		strBase = _strBaseDN;
		scope = LDAP_SCOPE_SUBTREE;
	}
	else
	{
		strBase = strSearch;
		scope = LDAP_SCOPE_BASE;
	}

// Find the user
	LDAPMessage* presult( NULL );
	LDAPMessage* psubresult( NULL );

#if defined (LINUX)
	int nRC = ldap_search_ext_s( pld,
									strBase.c_str(),
									scope,
									strWhat.length() ? strWhat.c_str() : NULL,
									NULL,
									0,
									NULL,
									NULL,
									NULL,
									100,
									&presult );
#elif defined (_WINDOWS)
	int nRC = ldap_search_sA( pld, (const PCHAR)strBase.c_str(), scope, (const PCHAR)( strWhat.length() ? strWhat.c_str() : NULL ), NULL, 0, &presult );
#endif /* LINUX */

	if( LDAP_SUCCESS != nRC  )
	{
		if( NULL != presult )
			ldap_msgfree( presult );

#if defined (LINUX)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2string( nRC ) );
#elif defined (_WINDOWS)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, ldap_err2stringA( nRC ) );
#endif /* LINUX */
	}

	if( NULL != presult )
	{
		if( 1 != ldap_count_entries( pld, presult ) )
		{
			ldap_msgfree( presult );

			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_NOUNIQUEUSER, "There is no unique user for specified user name." );
		}

		psubresult = ldap_first_entry( pld, presult );
		if( NULL != psubresult )
		{
#if defined (LINUX)
			strDNUser = ldap_get_dn( pld, psubresult );

			struct berval** pvalS;

			// userId
			pvalS = ldap_get_values_len( pld, psubresult, "samaccountname" );
			if( 0 < ldap_count_values_len( pvalS ) )
			{
				Auth4cData::User userL( domain(), pvalS[0]->bv_val );
				user = userL;
			}
			ldap_value_free_len( pvalS );

			// name
			std::string strName( "" );
			pvalS = ldap_get_values_len( pld, psubresult, "givenName" ); // first name
			if( 0 < ldap_count_values_len( pvalS ) )
			{
				strName = pvalS[0]->bv_val;
			}
			ldap_value_free_len( pvalS );

			pvalS = ldap_get_values_len( pld, psubresult, "sn" ); // last name
			if( 0 < ldap_count_values_len( pvalS ) )
			{
				if( "" != strName )
				{
					strName += " ";
				}
				strName += pvalS[0]->bv_val;
			}
			ldap_value_free_len( pvalS );
			
			user.setName( strName );

			// contact
			pvalS = ldap_get_values_len( pld, psubresult, "mail" );
			if( 0 < ldap_count_values_len( pvalS ) ) // "mail" is optional so this may not exist
			{
				user.setContact( pvalS[0]->bv_val );
			}
			ldap_value_free_len( pvalS );

#elif defined (_WINDOWS)
			strDNUser = ldap_get_dnA( pld, psubresult );

			char** pvalS;

			// userId
			pvalS = ldap_get_valuesA( pld, psubresult, "samaccountname" );
			if( 0 < ldap_count_valuesA( pvalS ) ) // "mail" is optional so this may not exist
			{
				Auth4cData::User userL( domain(), pvalS[0] );
				user = userL;
			}
			ldap_value_freeA( pvalS );

			// name
			std::string strName( "" );
			pvalS = ldap_get_valuesA( pld, psubresult, "givenName" ); // first name
			if( 0 < ldap_count_valuesA( pvalS ) )
			{
				strName = pvalS[0];
			}
			ldap_value_freeA( pvalS );

			pvalS = ldap_get_valuesA( pld, psubresult, "sn" ); // last name
			if( 0 < ldap_count_valuesA( pvalS ) )
			{
				if( "" != strName )
				{
					strName += " ";
				}
				strName += pvalS[0];
			}
			ldap_value_freeA( pvalS );

			user.setName( strName );

			// contact
			pvalS = ldap_get_valuesA( pld, psubresult, "mail" );
			if( 0 < ldap_count_valuesA( pvalS ) ) // "mail" is optional so this may not exist
			{
				user.setContact( pvalS[0] );
			}
			ldap_value_freeA( pvalS );

#endif /* LINUX */
		}

		ldap_msgfree( presult );
	}

	presult = NULL;

// Get Groups
	if( bTeams )
	{
		std::string strWho;
		std::string strDNGroup;
		std::vector<std::string> ldapGroupS;

		strWho = "member=";
		strWho += strDNUser;

#if defined (LINUX)
		nRC = ldap_search_ext_s( pld,
										_strBaseDN.c_str(),
										LDAP_SCOPE_SUBTREE,
										strWho.c_str(),
										NULL,
										0,
										NULL,
										NULL,
										NULL,
										100,
										&presult );
#elif defined (_WINDOWS)
		nRC = ldap_search_sA( pld, (const PCHAR)_strBaseDN.c_str(), LDAP_SCOPE_SUBTREE, (const PCHAR)strWho.c_str(), NULL, 0, &presult );
#endif /* LINUX */

		if( LDAP_SUCCESS != nRC  )
		{
		}
		else
		{
			psubresult = ldap_first_entry( pld, presult );
			while( NULL != psubresult )
			{
#if defined (LINUX)
				strDNGroup = ldap_get_dn( pld, psubresult );
#elif defined (_WINDOWS)
				strDNGroup = ldap_get_dnA( pld, psubresult );
#endif /* LINUX */

				ldapGroupS.push_back( strDNGroup );

				psubresult = ldap_next_entry( pld, psubresult );
			}
		}

		if( NULL != presult )
		{
			ldap_msgfree( presult );
		}

		presult = NULL;

		for( std::vector<std::string>::const_iterator iter = ldapGroupS.begin();
				iter != ldapGroupS.end();
				++iter )
		{
			Auth4cData::AlienKey key( std::make_pair( user.domain().domain(), *iter ) );
			const Auth4cData::XTeamPtrMap& xteamS( Auth4cData::Factory::teams( key ) );
			for( Auth4cData::XTeamPtrMap::const_iterator iterXTeam = xteamS.begin();
					iterXTeam != xteamS.end();
					++iterXTeam )
			{
				Auth4cData::TeamMap::const_iterator iterTeamReal = Auth4cData::Factory::teams().find( iterXTeam->second->team() );
				if( Auth4cData::Factory::teams().end() != iterTeamReal )
				{
					user.teams()[iterTeamReal->second.team()] = iterTeamReal->second;
				}
			}
		}
	}
}

void Authenticator::findUsers( Auth4cData::UserMap& userS, const std::string strSearch )
{
	throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_NOTIMPLEMENTED, "Feature not implemented" );
}

std::vector<std::string> Authenticator::groupMembers( const std::string& strGroup, LDAP* pld )
{
	std::vector<std::string> strRtnS;

	std::string strSearch( "memberOf=" );
	strSearch += strGroup;

	LDAPMessage* presult( NULL );
	LDAPMessage* psubresult( NULL );
	std::string strDNUser;
	std::string strDNGroup;

#if defined (LINUX)
	int nRC = ldap_search_ext_s( pld,
									_strBaseDN.c_str(),
									LDAP_SCOPE_SUBTREE,
									strSearch.c_str(),
									NULL,
									0,
									NULL,
									NULL,
									NULL,
									100,
									&presult );
#elif defined (_WINDOWS)
	int nRC = ldap_search_sA( pld, (const PCHAR)_strBaseDN.c_str(), LDAP_SCOPE_SUBTREE, (const PCHAR)strSearch.c_str(), NULL, 0, &presult );
#endif /* LINUX */

	if( LDAP_SUCCESS != nRC  )
	{
		if( NULL != presult )
		{
			ldap_msgfree( presult );
		}

		presult = NULL;

// Maybe throw something?
	}

	if( NULL != presult )
	{
		psubresult = ldap_first_entry( pld, presult );
		while( NULL != psubresult )
		{
#if defined (LINUX)
			strDNUser = ldap_get_dn( pld, psubresult );
#elif defined (_WINDOWS)
			strDNUser = ldap_get_dnA( pld, psubresult );
#endif /* LINUX */
			strRtnS.push_back( strDNUser );

			psubresult = ldap_next_entry( pld, psubresult );
		}

		ldap_msgfree( presult );
	}

	return strRtnS;
}


//////////////////////////////////////////////////////////////////////////
//
// AuthenticationLibrary
//

AuthenticationLibrary::AuthenticationLibrary()
{
}

AuthenticationLibrary::AuthenticationLibrary( const Auth4cData::AuthenticationLibrary* ptemplate )
	: Auth4cData::AuthenticationLibrary( ptemplate )
{
}

AuthenticationLibrary::AuthenticationLibrary( const AuthenticationLibrary& library )
	: Auth4cData::AuthenticationLibrary( library )
{
}

AuthenticationLibrary::~AuthenticationLibrary()
{
}

const AuthenticationLibrary& AuthenticationLibrary::operator=( const AuthenticationLibrary& library )
{
	if( this != &library )
	{
		Auth4cData::AuthenticationLibrary::operator=( library );
	}

	return *this;
}

Auth4cData::Authenticator* AuthenticationLibrary::authenticator( const Auth4cData::Domain& domain,
																	const std::string& strDescription,
																	const Auth4cData::InitMap& initMap )
{
	Authenticator* pauthenticator( NULL );
	pauthenticator = new Authenticator( initMap,
										domain,
										strDescription,
										authenticationLibrary() );

	return pauthenticator;
}
extern "C" Auth4cAD_Export Auth4cData::AuthenticationLibrary* library( const Auth4cData::AuthenticationLibrary* ptemplate )
{
	return new AuthenticationLibrary( ptemplate );
}
