/*

  $Id: Auth4cAuth4c.cpp 739 2020-09-30 15:36:00Z john $

  Copyright (c) 2008, The 4c Company. All rights reserved.
*/

#include <Auth4cAuth4c/Auth4cAuth4c.h>

#include <Auth4cDal/Auth4cDB.h>

#include <Auth4cDal/UserQuery.h>
#include <Auth4cDal/UserTeamQuery.h>
#include <Auth4cDal/DBProfileQuery.h>

#include <Auth4cData/Exception.h>
#include <Auth4cData/Domain.h>
#include <Auth4cData/AuthenticationToken.h>
#include <Auth4cData/SessionString.h>
#include <Auth4cData/InitMap.h>

#include <Auth4cDataDB/Auth4cFactory.h>

#include <librfcdal/SqlFunction.h>
#include <librfcdal/SqlInteger.h>
#include <librfcdal/SqlExpression.h>

#include <sstream>

RfcDal::SqlSchema g_schema;

using namespace Auth4cAuth4c;

//////////////////////////////////////////////////////////////////////////
//
// Authenticator
//

Authenticator::Authenticator( const Auth4cData::InitMap& initAuthenticator,
								const Auth4cData::Domain& domain,
								const std::string& strDescription,
								int nAuthenticationLibrary )
	: Auth4cData::Authenticator( initAuthenticator,
									domain,
									strDescription,
									nAuthenticationLibrary )
{
}

Authenticator::Authenticator( const Authenticator& authenticator )
	: Auth4cData::Authenticator( authenticator )
{
}

Authenticator::~Authenticator()
{
}

const Authenticator& Authenticator::operator=( const Authenticator& authenticator )
{
	if( this != &authenticator )
	{
		Auth4cData::Authenticator::operator=( authenticator );
	}

	return *this;
}

Auth4cData::AuthenticatedUser Authenticator::authenticate( const Auth4cData::User& user, const Auth4cData::AuthenticationToken& authToken, const Auth4cData::Profile& profile )
{
	RfcDal::DalMySQLDatabase db;
	init( db );
	return authenticate( user, authToken, profile, db );
}

void Authenticator::user( Auth4cData::User& userF, bool bRecurse /*= true*/ )
{
	RfcDal::DalMySQLDatabase db;
	init( db );
	user( userF, db, bRecurse );
}

void Authenticator::team( Auth4cData::Team& teamF )
{
	RfcDal::DalMySQLDatabase db;
	init( db );
	team( teamF, db );
}

void Authenticator::findUsers( Auth4cData::UserMap& userS, const std::string strSearch )
{
	RfcDal::DalMySQLDatabase db;
	init( db );
	findUsers( userS, strSearch, db );
}

void Authenticator::users( Auth4cData::UserMap& userS )
{
	RfcDal::DalMySQLDatabase db;
	init( db );

	RfcDal::DalDatabase::Result result;
	RfcDal::DalDatabase::Row row;

	Auth4cDal::UserQuery qryUser( Auth4cDal::schema() );

	qryUser.constrainDomain( RfcDal::SqlText( domain().domain() ) );
	qryUser.select( result, db );
	for( RfcDal::DalDatabase::Result::const_iterator iter = result.begin();
			iter != result.end();
			++iter )
	{
		row = *iter;
		Auth4cData::User user( Auth4cDataDB::Factory::userFromRow( row ) );
		userS[user.key()] = user;
	}
}

void Authenticator::passwd( const Auth4cData::AuthenticatedUser& user, Auth4cData::AuthenticationToken& authTokenCurrent, Auth4cData::AuthenticationToken& authTokenNew )
{
	RfcDal::DalMySQLDatabase db;
	init( db );

	Auth4cDal::UserQuery qry( g_schema );
	qry.constrainDomain( RfcDal::SqlText( user.domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( user.userId() ) );

	RfcDal::SqlText txtAuthenticationToken( authTokenCurrent.str() );
	RfcDal::SqlFunction fnPassword( "PASSWORD" );
	fnPassword.addColumn( txtAuthenticationToken );

	RfcDal::DalDatabase::Result result;
	RfcDal::DalDatabase::Row row;

	if( !qry.select( result, db )  )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Query did not return exactly 1 result." );
	}

	RfcDal::SqlText txtAuthenticationTokenNew( authTokenNew.str() );
	RfcDal::SqlFunction fnPasswordNew( "PASSWORD" );
	fnPasswordNew.addColumn( txtAuthenticationTokenNew );

	qry.columnAuthenticationToken().setValue( fnPasswordNew );
	qry.update( db );
}

void Authenticator::createUser( const Auth4cData::User& user )
{
	RfcDal::DalMySQLDatabase db;
	init( db );

	Auth4cDal::UserQuery qry( g_schema );

	if( domain().domain() != user.domain().domain() )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DOMAINNOTFOUND, user.domain().domain(), true );
	}

	if( user.authenticationToken().length() )
	{
		RfcDal::SqlText txtAuthenticationTokenNew( user.authenticationToken() );
		RfcDal::SqlFunction fnPasswordNew( "PASSWORD" );
		fnPasswordNew.addColumn( txtAuthenticationTokenNew );
		qry.columnAuthenticationToken().setValue( fnPasswordNew );
	}
	else
	{
		qry.columnAuthenticationToken().setValue( RfcDal::SqlText( "" ) );
	}

	qry.columnDomain().setValue( RfcDal::SqlText( user.domain().domain() ) );
	qry.columnUserId().setValue( RfcDal::SqlText( user.userId() ) );
	qry.columnName().setValue( RfcDal::SqlText( user.name() ) );
	qry.columnContact().setValue( RfcDal::SqlText( user.contact() ) );
	qry.columnStatus().setValue( RfcDal::SqlText( "1" ) );
	qry.columnBadLoginCount().setValue( RfcDal::SqlInteger( 0 ) );

	try
	{
		qry.insert( db );
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, e.nativeExplanation(), true );
	}
}

void Authenticator::updateUser( const Auth4cData::User& user )
{
	RfcDal::DalMySQLDatabase db;
	init( db );

	Auth4cDal::UserQuery qry( g_schema );

	if( domain().domain() != user.domain().domain() )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DOMAINNOTFOUND, user.domain().domain(), true );
	}

	qry.constrainDomain( RfcDal::SqlText( user.domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( user.userId() ) );

	if( user.authenticationToken().length() )
	{
		RfcDal::SqlText txtAuthenticationTokenNew( user.authenticationToken() );
		RfcDal::SqlFunction fnPasswordNew( "PASSWORD" );
		fnPasswordNew.addColumn( txtAuthenticationTokenNew );
		qry.columnAuthenticationToken().setValue( fnPasswordNew );
	}

	qry.columnName().setValue( RfcDal::SqlText( user.name() ) );
	qry.columnContact().setValue( RfcDal::SqlText( user.contact() ) );
	qry.columnStatus().setValue( RfcDal::SqlText( "1" ) );
	qry.columnBadLoginCount().setValue( RfcDal::SqlInteger( 0 ) );

	try
	{
		if( 1 != qry.update( db ) )
		{
//			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, "No user was updated", true );
		}
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, e.nativeExplanation(), true );
	}
}

void Authenticator::deleteUser( const Auth4cData::User& user )
{
	RfcDal::DalMySQLDatabase db;
	init( db );

	Auth4cDal::UserQuery qry( g_schema );

	if( domain().domain() != user.domain().domain() )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DOMAINNOTFOUND, user.domain().domain(), true );
	}

	qry.constrainDomain( RfcDal::SqlText( user.domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( user.userId() ) );

	try
	{
		if( 1 != qry.remove( db ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, "No user was deleted", true );
		}
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, e.nativeExplanation(), true );
	}
}

void Authenticator::createTeamMember( Auth4cData::Team& team, Auth4cData::User& user )
{
	RfcDal::DalMySQLDatabase db;
	init( db );

	Auth4cDal::UserTeamQuery qry( g_schema );

	if( domain().domain() != user.domain().domain() )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DOMAINNOTFOUND, user.domain().domain(), true );
	}

	qry.columnDomain().setValue( RfcDal::SqlText( user.domain().domain() ) );
	qry.columnUserId().setValue( RfcDal::SqlText( user.userId() ) );
	qry.columnTeam().setValue( RfcDal::SqlInteger( team.team() ) );
	qry.columnIsLead().setValue( RfcDal::SqlInteger( 0 ) );

	try
	{
		qry.insert( db );

		user.teams()[team.team()] = team;
		Auth4cData::TeamMember member( user, false );
		team.members()[member.key()] = member;
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, e.nativeExplanation(), true );
	}
}

void Authenticator::deleteTeamMember( Auth4cData::Team& team, Auth4cData::User& user )
{
	RfcDal::DalMySQLDatabase db;
	init( db );

	Auth4cDal::UserTeamQuery qry( g_schema );

	if( domain().domain() != user.domain().domain() )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DOMAINNOTFOUND, user.domain().domain(), true );
	}

	qry.constrainDomain( RfcDal::SqlText( user.domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( user.userId() ) );
	qry.constrainTeam( RfcDal::SqlInteger( team.team() ) );

	try
	{
		if( 1 != qry.remove( db ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, "No member was deleted", true );
		}

		Auth4cData::TeamMemberMap::iterator iterMember = team.members().find( user.key() );
		if( team.members().end() != iterMember )
		{
			team.members().erase( iterMember );
		}

		Auth4cData::TeamMap::iterator iterTeam = user.teams().find( team.team() );
		if( user.teams().end() != iterTeam )
		{
			user.teams().erase( iterTeam );
		}
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_DATABASE, e.nativeExplanation(), true );
	}
}

Auth4cData::AuthenticatedUser Authenticator::authenticate( const Auth4cData::User& user, const Auth4cData::AuthenticationToken& authToken, const Auth4cData::Profile& profile, RfcDal::DalDatabase& db )
{

	Auth4cDal::UserQuery qry( g_schema );
	qry.constrainDomain( RfcDal::SqlText( user.domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( user.userId() ) );

	RfcDal::SqlText txtAuthenticationToken( authToken.str() );
	RfcDal::SqlFunction fnPassword( "PASSWORD" );
	fnPassword.addColumn( txtAuthenticationToken );

	RfcDal::DalDatabase::Result result;
	RfcDal::DalDatabase::Row row;
	
	Auth4cDal::DBProfileQuery qryDbp( g_schema );
	qryDbp.constrainPName( RfcDal::SqlText( "MaxBadLoginCount" ) );
	
	RfcDal::DalDatabase::Result resultDbp;
	RfcDal::DalDatabase::Row rowDbp;

	if ( qry.select( result, db ) > 0 )
	{
		row = *( result.begin() );
		RfcDal::DalDatabase::Row::const_iterator column;
		column = row.find( qry.columnBadLoginCount().id() );
		if( row.end() != column )
		{
			int nBadLoginCount = column->second.asInt();

			if( qryDbp.select( resultDbp, db ) )
			{
				rowDbp = *( resultDbp.begin() );

				RfcDal::DalDatabase::Row::const_iterator columnDbp;
				columnDbp = rowDbp.find( qryDbp.columnPValue().id() );
				if( rowDbp.end() != columnDbp )
				{
					int nMaxLogins = columnDbp->second.asInt();
					if( nBadLoginCount < nMaxLogins ) 
					{
						qry.constrainAuthenticationToken( fnPassword );
						if( 1 != qry.select( result, db ) )
						{
							incrementBadLogins( user, db );
							throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Query did not return exactly 1 result." );
						}
						else
						{
							Auth4cData::User userA( user );
							this->user( userA, db );
						
							std::ostringstream os;
							os << &userA;						
							Auth4cData::SessionString sesStr( os.str() );
							Auth4cData::AuthenticatedUser userAuth( userA, sesStr, profile );
							resetBadLogins( userA, db );
							return userAuth;
						}
					}
				}
			}
			else
			{
				throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Max login count has not been correctly set in DBProfile. cf 'MaxBadLoginCount'." );
			}
		}
	}
	throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Query did not return exactly 1 result." );
}

void Authenticator::user( Auth4cData::User& userF, RfcDal::DalDatabase& db, bool bRecurse /*= true*/ )
{
	user( userF, userF.userId(), db );
}

void Authenticator::team( Auth4cData::Team& team, RfcDal::DalDatabase& db )
{
	Auth4cDal::UserTeamQuery qry( g_schema );
	qry.constrainDomain( RfcDal::SqlText( this->domain().domain() ) );
	qry.constrainTeam( RfcDal::SqlInteger( team.team() ) );

	RfcDal::DalDatabase::Result result;
	RfcDal::DalDatabase::Row row;
	
	qry.select( result, db );
//	row = *( result.begin() );
	RfcDal::DalDatabase::Row::const_iterator column;

	std::string strUserMember;
	int nLead;

	for( RfcDal::DalDatabase::Result::const_iterator iter = result.begin();
		iter != result.end();
		++iter )
	{
		row = *iter;

		column = row.find( qry.columnUserId().id() );
		if( row.end() != column )
		{
			strUserMember = column->second.string();
		}

		column = row.find( qry.columnIsLead().id() );
		if( row.end() != column )
		{
			nLead = column->second.asInt();
		}

		try
		{
			Auth4cData::User userF;
			user( userF, strUserMember, db );
			Auth4cData::TeamMember member( userF, ( nLead ? true : false ) );
			team.members()[member.key()] = member;
		}
		catch( ... )
		{
			// only happens when the user is non-existent
		}
	}
}

void Authenticator::user( Auth4cData::User& user, const std::string& strSearch, RfcDal::DalDatabase& db )
{
	Auth4cDal::UserQuery qry( g_schema );
	qry.constrainDomain( RfcDal::SqlText( domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( strSearch ) );

	Auth4cDal::UserTeamQuery qryUT( g_schema );
	qryUT.constrainDomain( RfcDal::SqlText( domain().domain() ) );
	qryUT.constrainUserId( RfcDal::SqlText( strSearch ) );

	RfcDal::DalDatabase::Result result;
	RfcDal::DalDatabase::Row row;
	if( 1 != qry.select( result, db ) )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, "Query did not return exactly 1 result." );
	}
	else
	{
		row = *( result.begin() );
		RfcDal::DalDatabase::Row::const_iterator column;

		column = row.find( qry.columnUserId().id() );
		Auth4cData::User userL( domain(), column->second.string() );

		column = row.find( qry.columnName().id() );
		if( row.end() != column )
		{
			userL.setName( column->second.string() );
		}

		column = row.find( qry.columnAuthenticationToken().id() );
		if( row.end() != column )
		{
			userL.setAuthenticationToken( column->second.string() );
		}

		column = row.find( qry.columnContact().id() );
		if( row.end() != column )
		{
			userL.setContact( column->second.string() );
		}

		result.clear();
		qryUT.select( result, db );
		for( RfcDal::DalDatabase::Result::const_iterator itTeamR = result.begin();
			itTeamR != result.end();
			++itTeamR )
		{
			const RfcDal::DalDatabase::Row& row = *itTeamR;
			RfcDal::DalDatabase::Row::const_iterator colTeam = row.find( qryUT.columnTeam().id() );
			if( row.end() != colTeam )
			{
				int nTeam = colTeam->second.asInt();
				Auth4cData::TeamMap::iterator itTeam = Auth4cData::Factory::teams().find( nTeam );
				if( Auth4cData::Factory::teams().end() != itTeam )
				{
					userL.teams()[itTeam->first] = itTeam->second;
				}
			}
		}

		user = userL;
	}
}

void Authenticator::findUsers( Auth4cData::UserMap& userS, const std::string strSearch, RfcDal::DalDatabase& db )
{
	Auth4cDal::UserQuery qry( g_schema );

	qry.constrainDomain( RfcDal::SqlText( domain().domain() ) );
	qry.constrainSearch( RfcDal::SqlValue( strSearch ) );

	RfcDal::DalDatabase::Result result;
	RfcDal::DalDatabase::Row row;

	qry.select( result, db );
	RfcDal::DalDatabase::Row::const_iterator column;

	for( RfcDal::DalDatabase::Result::const_iterator iter = result.begin();
		iter != result.end();
		++iter )
	{
		row = *iter;
		column = row.find( qry.columnUserId().id() );

		if( row.end() != column )
		{
			Auth4cData::User userF;
			user( userF, column->second.string(), db );
			userS[userF.key()] = userF;
		}
	}
}

void Authenticator::init( RfcDal::DalDatabase& db )
{
	try
	{
		std::string strXConnect;

		Auth4cData::InitMap::const_iterator iterXConnect = authenticatorInit().find( "XConnect" );
		if( authenticatorInit().end() != iterXConnect )
		{
			strXConnect = iterXConnect->second;
		}

		RfcDal::DalDatabase::ConnectionParameterMap map;
		db.connectStringToMap( map, strXConnect );
		
		RfcDal::DalDatabase::ConnectionParameterMap::const_iterator iter = map.find( "database" );
		if( iter != map.end() )
		{
			try
			{
				db.connectX( map );
			}
			catch( const RfcDal::DalDatabaseException& e )
			{
				throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATORINITFAILED, e.nativeExplanation() );
			}

			RfcDal::SqlSchema schema( iter->second );
			g_schema = schema;
		}
		else
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATORINITFAILED, "No schema/database." );
		}
	}
	catch( const Auth4cData::Exception& e )
	{
		throw;
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATORINITFAILED, e.nativeExplanation() );
	}
}

void Authenticator::incrementBadLogins( const Auth4cData::User& user, RfcDal::DalDatabase& db  )
{
	Auth4cDal::UserQuery qry( g_schema );
	qry.constrainDomain( RfcDal::SqlText( user.domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( user.userId() ) );

	qry.columnBadLoginCount().setValue( RfcDal::SqlExpression( qry.columnBadLoginCount(), RfcDal::SqlInteger(1), RfcDal::SqlExpression::expPlus ) );

	try
	{
		if ( 1 != qry.update( db ) )
		{
			// @TODO: raise a warning
		}
	}
	catch ( ... )
	{
		// update error
	}
}

void Authenticator::resetBadLogins( const Auth4cData::User& user, RfcDal::DalDatabase& db  )
{
	Auth4cDal::UserQuery qry( g_schema );
	qry.constrainDomain( RfcDal::SqlText( user.domain().domain() ) );
	qry.constrainUserId( RfcDal::SqlText( user.userId() ) );

	qry.columnBadLoginCount().setValue( RfcDal::SqlInteger(0) );

	try
	{
		if ( 1 != qry.update( db ) )
		{
			// @TODO: raise a warning
		}
	}
	catch ( ... )
	{
		// update error
	}
}


//////////////////////////////////////////////////////////////////////////
//
// AuthenticationLibrary
//

AuthenticationLibrary::AuthenticationLibrary()
{
}

AuthenticationLibrary::AuthenticationLibrary( const Auth4cData::AuthenticationLibrary* ptemplate )
	: Auth4cData::AuthenticationLibrary( ptemplate )
{
}

AuthenticationLibrary::AuthenticationLibrary( const AuthenticationLibrary& library )
	: Auth4cData::AuthenticationLibrary( library )
{
}

AuthenticationLibrary::~AuthenticationLibrary()
{
}

const AuthenticationLibrary& AuthenticationLibrary::operator=( const AuthenticationLibrary& library )
{
	if( this != &library )
	{
		Auth4cData::AuthenticationLibrary::operator=( library );
	}

	return *this;
}

Auth4cData::Authenticator* AuthenticationLibrary::authenticator( const Auth4cData::Domain& domain,
																	const std::string& strDescription,
																	const Auth4cData::InitMap& initMap )
{
	Authenticator* pauthenticator = new Authenticator( initMap,
														domain,
														strDescription,
														authenticationLibrary() );

	return pauthenticator;
}

extern "C" Auth4cAuth4c_Export Auth4cData::AuthenticationLibrary* library( const Auth4cData::AuthenticationLibrary* ptemplate )
{
	return new AuthenticationLibrary( ptemplate );
}
