/*

  $Id: AuthenticationLibrary.cpp 689 2018-11-28 17:54:35Z john $

  Copyright (c) 2008, The 4c Company. All rights reserved.
*/

#include <Auth4cData/AuthenticationLibrary.h>
#include <Auth4cData/Factory.h>

#if defined (LINUX)
#	include <dlfcn.h>
#endif /* LINUX */

using namespace Auth4cData;

AuthenticationLibrary::AuthenticationLibrary()
	: _nAuthenticationLibrary( -1 ),
	_authenticationTokenDisposition( dispositionUnknown ),
	_bUserTools( false ),
	_handle( NULL )
{
}

AuthenticationLibrary::AuthenticationLibrary( const AuthenticationLibrary* ptemplate )
	: _nAuthenticationLibrary( ptemplate->_nAuthenticationLibrary ),
	_initLibrary( ptemplate->_initLibrary ),
	_authenticationTokenDisposition( ptemplate->_authenticationTokenDisposition ),
	_bUserTools( ptemplate->_bUserTools ),
	_sharedObjectS( ptemplate->_sharedObjectS ),
	_handle( NULL ) // NOT _handle = ptemplate->_handle as this will lead to a double free
{
}

AuthenticationLibrary::AuthenticationLibrary( int nAuthenticationLibrary,
												const std::string& strDescription,
												const InitMap& initLibrary,
												AuthenticationTokenDisposition authenticationTokenDisposition,
												bool bUserTools )
	: _nAuthenticationLibrary( nAuthenticationLibrary ),
	_strDescription( strDescription ),
	_initLibrary( initLibrary ),
	_authenticationTokenDisposition( authenticationTokenDisposition ),
	_bUserTools( bUserTools ),
	_handle( NULL )
{
}

AuthenticationLibrary::AuthenticationLibrary( const AuthenticationLibrary& library )
	: _nAuthenticationLibrary( library._nAuthenticationLibrary ),
	_strDescription( library._strDescription ),
	_initLibrary( library._initLibrary ),
	_authenticationTokenDisposition( library._authenticationTokenDisposition ),
	_bUserTools( library._bUserTools ),
	_sharedObjectS( library._sharedObjectS ),
	_handle( library._handle )
{
}

AuthenticationLibrary::~AuthenticationLibrary()
{
	if( _handle != NULL )
	{
#if defined (LINUX)
		dlclose( _handle );
#elif defined (_WINDOWS)
		FreeLibrary( _handle );
#endif /* LINUX */
	}
}

const AuthenticationLibrary& AuthenticationLibrary::operator=( const AuthenticationLibrary& library )
{
	if( this != &library )
	{
		_nAuthenticationLibrary = library._nAuthenticationLibrary;
		_strDescription = library._strDescription;
		_initLibrary = library._initLibrary;
		_authenticationTokenDisposition = library._authenticationTokenDisposition;
		_bUserTools = library._bUserTools;
		_sharedObjectS = library._sharedObjectS;
		_handle = library._handle;
	}

	return *this;
}

AuthenticationLibrary* AuthenticationLibrary::load()
{
	AuthenticationLibrary* plibrary = NULL;

	if( _handle == NULL )
	{
#if defined (LINUX)
		_handle = dlopen(  sharedObject().sharedObject().c_str(), RTLD_LAZY );
#elif defined (_WINDOWS)
		_handle = LoadLibraryA( sharedObject().sharedObject().c_str() );
#endif /* LINUX */
	}

	if( _handle == NULL )
	{
#if defined (LINUX)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_NOSHAREDOBJECT, dlerror() );
#elif defined (_WINDOWS)
		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_NOSHAREDOBJECT, "Could not load shared object" );
#endif /* LINUX */
	}
	else
	{
		Auth4cData::AuthenticationLibrary* (*library)( const AuthenticationLibrary* ) = NULL;
#if defined (LINUX)
		library = (Auth4cData::AuthenticationLibrary* (*)( const AuthenticationLibrary* ))dlsym( _handle, "library" );
#elif defined (_WINDOWS)
		library = (Auth4cData::AuthenticationLibrary* (*)( const AuthenticationLibrary* ))GetProcAddress( _handle, "library" );
#endif /* LINUX */
			
#if defined (LINUX)
		char* pszError = dlerror();
		if( pszError != NULL )
		{
			dlclose( _handle );
			_handle = NULL;
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_NOENTRYPOINT, "Symbol not found" );
		}
#elif defined (_WINDOWS)
		if( library == NULL )
		{
			FreeLibrary( _handle );
			_handle = NULL;
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_NOENTRYPOINT, "Symbol not found" );
		}
#endif /* LINUX */

		plibrary = library( this );

		Factory::authenticationLibraries()[_nAuthenticationLibrary] = plibrary;

		return plibrary;
	}

	return NULL;
}

int AuthenticationLibrary::authenticationLibrary() const
{
	return _nAuthenticationLibrary;
}

const std::string& AuthenticationLibrary::description() const
{
	return _strDescription;
}

//const std::string& AuthenticationLibrary::sharedObject() const

const SharedObject& AuthenticationLibrary::sharedObject( const std::string& strPlatform /*= std::string()*/ ) const
{
	std::string strPlatformL( strPlatform );

	if( !strPlatformL.length() )
	{
#if defined (LINUX)
		strPlatformL = "LNX";
#elif defined (_WINDOWS)
		strPlatformL = "WIN";
#endif /* LINUX */
	}

	Auth4cData::SharedObjectPtrMap::const_iterator iter = _sharedObjectS.find( strPlatformL );

	if( _sharedObjectS.end() == iter )
	{
		throw Exception( __FILE__, __LINE__, nNAX_NOSHAREDOBJECTSET, strPlatformL );
	}

	return *iter->second;
}

const InitMap& AuthenticationLibrary::libraryInit() const
{
	return _initLibrary;
}

AuthenticationLibrary::AuthenticationTokenDisposition AuthenticationLibrary::authenticationTokenDisposition() const
{
	return _authenticationTokenDisposition;
}

bool AuthenticationLibrary::userTools() const
{
	return _bUserTools;
}

const SharedObjectPtrMap& AuthenticationLibrary::sharedObjects() const
{
	return _sharedObjectS;
}

void AuthenticationLibrary::setDescription( const std::string& strDescription )
{
	_strDescription = strDescription;
}

void AuthenticationLibrary::setLibraryInit( const InitMap& initLibrary )
{
	_initLibrary = initLibrary;
}

void AuthenticationLibrary::setAuthenticationTokenDisposition( AuthenticationLibrary::AuthenticationTokenDisposition authenticationTokenDisposition )
{
	_authenticationTokenDisposition = authenticationTokenDisposition;
}

void AuthenticationLibrary::setUserTools( bool bUserTools )
{
	_bUserTools = bUserTools;
}

SharedObjectPtrMap& AuthenticationLibrary::sharedObjects()
{
	return _sharedObjectS;
}

Authenticator* AuthenticationLibrary::authenticator( const Domain& domain,
														const std::string& strDescription,
														const InitMap& initAuthenticator )
{
	return NULL;
}
