/*

  $Id: User.cpp 689 2018-11-28 17:54:35Z john $

  Copyright (c) 2008, The 4c Company. All rights reserved.
*/

#include <Auth4cData/User.h>

#include <Auth4cData/Team.h>
#include <Auth4cData/Factory.h>

#include <Base4c/Util.h>

using namespace Auth4cData;

User::User()
	: _pdomain( NULL )
{
	_pteamS = new TeamMap;
}

User::User( const Domain& domain,
			const std::string& strUserId,
			const std::string& strName,
			const std::string& strAuthenticationToken,
			const std::string& strContact )
	: _pdomain( &domain ),
	_strUserId( strUserId ),
	_strName( strName ),
	_strAuthenticationToken( strAuthenticationToken ),
	_strContact( strContact ),
	_key( std::make_pair( _pdomain->domain(), _strUserId ) )
{
	_pteamS = new TeamMap;
}

User::User( const User& user )
	: _pdomain( user._pdomain ),
	_strUserId( user._strUserId ),
	_strName( user._strName ),
	_strAuthenticationToken( user._strAuthenticationToken ),
	_strContact( user._strContact ),
	_productS( user._productS ),
	_key( user._key )
{
	_pteamS = new TeamMap;
	_pteamS->operator=( *user._pteamS );
}

User::~User()
{
	if( NULL != _pteamS )
	{
		delete _pteamS;
		_pteamS = NULL;
	}
}

const User& User::operator=( const User& user )
{
	if( this != &user )
	{
		_pdomain = user._pdomain;
		_strUserId = user._strUserId;
		_strName = user._strName;
		_strAuthenticationToken = user._strAuthenticationToken;
		_strContact = user._strContact;
		_productS = user._productS;
		_pteamS->operator=( *user._pteamS );
		_key = user._key;
	}

	return *this;
}

bool User::operator==( const User& user ) const
{
	if( ( domain() == user.domain() ) &&
		( _strUserId == user._strUserId ) )
	{
		return true;
	}

	return false;
}

const Domain& User::domain() const
{
	if( NULL == _pdomain )
	{
		throw Exception( __FILE__, __LINE__, nNAX_NODOMAINSET, "Domain pointer is null." );
	}

	return *_pdomain;
}

const std::string& User::userId() const
{
	return _strUserId;
}

const std::string& User::name() const
{
	return _strName;
}

const std::string& User::authenticationToken() const
{
	return _strAuthenticationToken;
}

const std::string& User::contact() const
{
	return _strContact;
}

const ProductMap& User::products() const
{
	return _productS;
}

const TeamMap& User::teams() const
{
	return *_pteamS;
}

const User::Key& User::key() const
{
	return _key;
}

void User::setUserId( const std::string& strUserId )
{
	_strUserId = strUserId;
}

void User::setName( const std::string& strName )
{
	_strName = strName;
}

void User::setAuthenticationToken( const std::string& strAuthenticationToken )
{
	_strAuthenticationToken = strAuthenticationToken;
}

void User::setContact( const std::string& strContact )
{
	_strContact = strContact;
}

ProductMap& User::products()
{
	return _productS;
}

TeamMap& User::teams()
{
	return *_pteamS;
}

void User::update()
{
	domain().authenticator().updateUser( *this );
}

void User::addTeam( Team& team )
{
	domain().authenticator().createTeamMember( team, *this );
	_pteamS->operator[]( team.team() ) = team;
}

void User::deleteTeam( Team& team )
{
	domain().authenticator().deleteTeamMember( team, *this );
	TeamMap::iterator iter = _pteamS->find( team.team() );
	if( _pteamS->end() != iter )
	{
		_pteamS->erase( iter );
	}
}

bool User::hasRole( const Role& role ) const
{
	ProductMap::const_iterator iter = _productS.find( role.product().product() );
	if( _productS.end() == iter )
	{
		return false;
	}

	return iter->second.hasRole( role );
}

bool User::hasPermission( const Permission& permission ) const
{
	ProductMap::const_iterator iter = _productS.find( permission.product().product() );
	if( _productS.end() == iter )
	{
		return false;
	}

	return iter->second.hasPermission( permission );
}

bool User::isMemberOf( const Team& team ) const
{
	TeamMap::const_iterator iter = _pteamS->find( team.team() );
	if( _pteamS->end() == iter )
	{
		return false;
	}

	return true;
}

bool User::isMemberOf( const TeamMap& teamS ) const
{
	for( TeamMap::const_iterator iter = teamS.begin();
			iter != teamS.end();
			++iter )
	{
		if( isMemberOf( iter->second ) )
		{
			return true;
		}
	}

	return false;
}

User User::fromText( const std::string& strText )
{
	std::vector<std::string> strPartS;
	std::string strDomain;
	std::string strUserId;

	if( std::string::npos != strText.find( "/" ) )
	{
		// domain/user
		strPartS = Base4c::tokenize( strText, "/" );
		strDomain = strPartS[0];
		strUserId = strPartS[1];
	}
	else if( std::string::npos != strText.find( "\\" ) )
	{
		// domain\user
		strPartS = Base4c::tokenize( strText, "\\" );
		strDomain = strPartS[0];
		strUserId = strPartS[1];
	}
	else if( std::string::npos != strText.find( "@" ) )
	{
		// user@domain
		strPartS = Base4c::tokenize( strText, "@" );
		strDomain = strPartS[1];
		strUserId = strPartS[0];
	}
	else
	{
		throw Exception( __FILE__, __LINE__, nNAX_BADUSERFORMAT, "Expected one of 'domain/userid', 'domain\\userid' or 'userid@domain'" );
	}

	if( ( !strDomain.length() ) ||
		( !strUserId.length() ) )
	{
		throw Exception( __FILE__, __LINE__, nNAX_BADUSERFORMAT, "Expected one of 'domain/userid', 'domain\\userid' or 'userid@domain'" );
	}

	DomainMap::const_iterator iterDomain = Auth4cData::Factory::domains().find( strDomain );
	if( Auth4cData::Factory::domains().end() == iterDomain )
	{
		throw Exception( __FILE__, __LINE__, nNAX_DOMAINNOTFOUND, strDomain );
	}
	
	User user( iterDomain->second, strUserId );

	return user;
}
