/*

  $Id: TreeWidgetItem.cpp 689 2018-11-28 17:54:35Z john $

  Copyright (c) 2008, The 4c Company. All rights reserved.
*/

#include <Auth4cUIComponents/TreeWidgetItem.h>

#include <Auth4cData/Factory.h>
#include <Auth4cDataXML/Auth4cFactory.h>

#include <QMenu>
#include <QCursor>

using namespace Auth4cUIComponents;

//////////////////////////////////////////////////////////////////////////
//
// TreeWidgetItem
//

TreeWidgetItem::TreeWidgetItem( QTreeWidget* pparent, Auth4cData::Base* pbase, CheckState state /*= Uncheckable*/ )
	: QTreeWidgetItem( pparent ),
	_pbase( pbase ),
	_state( state )
{
	switch( _state )
	{
		case Unchecked:
			setCheckState( 0, Qt::Unchecked );
			break;

		case PartiallyChecked:
			setCheckState( 0, Qt::PartiallyChecked );
			break;

		case Checked:
			setCheckState( 0, Qt::Checked );
			break;
			
		case Uncheckable:
			break;
	}
}

TreeWidgetItem::TreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::Base* pbase, CheckState state /*= Uncheckable*/ )
	: QTreeWidgetItem( pparent ),
	_pbase( pbase ),
	_state( state )
{
	switch( _state )
	{
		case Uncheckable:
			break;

		case Unchecked:
			setCheckState( 0, Qt::Unchecked );
			break;

		case PartiallyChecked:
			setCheckState( 0, Qt::PartiallyChecked );
			break;

		case Checked:
			setCheckState( 0, Qt::Checked );
			break;
	}
}

TreeWidgetItem::~TreeWidgetItem()
{
}

TreeWidgetItem::CheckState TreeWidgetItem::state( bool bForReal /*= true*/ ) const
{
	if( Uncheckable == _state )
	{
		return Uncheckable;
	}

	if( bForReal )
	{
		TreeWidgetItem::CheckState state( Uncheckable );

		switch( checkState( 0 ) )
		{
			case Qt::Unchecked:
				state = Unchecked;
				break;

			case Qt::PartiallyChecked:
				state = PartiallyChecked;
				break;

			case Qt::Checked:
				state = Checked;
				break;
		}

		return state;
	}
	else
	{
		return _state;
	}
}

void TreeWidgetItem::setState( CheckState state, bool bForReal /*= true*/ )
{
	if( Uncheckable == _state )
	{
		return;
	}

	_state = state;

	if( bForReal )
	{
		switch( _state )
		{
			case Uncheckable:
				break;

			case Unchecked:
				setCheckState( 0, Qt::Unchecked );
				break;

			case PartiallyChecked:
				setCheckState( 0, Qt::PartiallyChecked );
				break;

			case Checked:
				setCheckState( 0, Qt::Checked );
				break;
		}
	}
}

Auth4cData::Base* TreeWidgetItem::base()
{
	return _pbase;
}

void TreeWidgetItem::refresh()
{
}

void TreeWidgetItem::doMenu()
{
}

void TreeWidgetItem::expanded()
{
}


//////////////////////////////////////////////////////////////////////////
//
// ProductTreeWidgetItem
//

ProductTreeWidgetItem::ProductTreeWidgetItem( QTreeWidget* pparent, Auth4cData::Product* pproduct, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pproduct, state ),
	_pproduct( pproduct )
{
	refresh();
}

ProductTreeWidgetItem::~ProductTreeWidgetItem()
{
}

void ProductTreeWidgetItem::refresh()
{
	this->setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/product.xpm" ) );
	setText( 0, QString::fromStdString( _pproduct->description() ) );

	new RoleRootTreeWidgetItem( this, _pproduct );

	new PermissionRootTreeWidgetItem( this, _pproduct );
}

void ProductTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// PermissionRootTreeWidgetItem
//

PermissionRootTreeWidgetItem::PermissionRootTreeWidgetItem( TreeWidgetItem* pparent,  Auth4cData::Product* pproduct, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pproduct, state ),
	_pproduct( pproduct )
{
	refresh();
}

PermissionRootTreeWidgetItem::~PermissionRootTreeWidgetItem()
{
}

void PermissionRootTreeWidgetItem::refresh()
{
	setText( 0, QObject::tr( "Permissions" ) );

	Auth4cData::PermissionMap& permissionS( _pproduct->permissions() );
	for( Auth4cData::PermissionMap::iterator iter = permissionS.begin();
			iter != permissionS.end();
			++iter )
	{
		new PermissionTreeWidgetItem( this, &iter->second );
	}
}

void PermissionRootTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// PermissionTreeWidgetItem
//

PermissionTreeWidgetItem::PermissionTreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::Permission* ppermission, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, ppermission, state ),
	_ppermission( ppermission )
{
	refresh();
}

PermissionTreeWidgetItem::~PermissionTreeWidgetItem()
{
}

void PermissionTreeWidgetItem::refresh()
{
	this->setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/perm_16.xpm" ) );
	setText( 0, QString::fromStdString( _ppermission->description() ) );
}

void PermissionTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// RoleRootTreeWidgetItem
//

RoleRootTreeWidgetItem::RoleRootTreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::Product* pproduct, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pproduct, state ),
	_pproduct( pproduct )
{
	refresh();
}

RoleRootTreeWidgetItem::~RoleRootTreeWidgetItem()
{
}

void RoleRootTreeWidgetItem::refresh()
{
	setText( 0, QObject::tr( "Roles" ) );

	Auth4cData::RoleMap& roleS( _pproduct->roles() );
	for( Auth4cData::RoleMap::iterator iter = roleS.begin();
			iter != roleS.end();
			++iter )
	{
		new RoleTreeWidgetItem( this, &iter->second );
	}
}

void RoleRootTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// RoleTreeWidgetItem
//

RoleTreeWidgetItem::RoleTreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::Role* prole, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, prole, state ),
	_prole( prole )
{
	refresh();
}

RoleTreeWidgetItem::~RoleTreeWidgetItem()
{
}

void RoleTreeWidgetItem::refresh()
{
	this->setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/role_16.xpm" ) );
	setText( 0, QString::fromStdString( _prole->description() ) );

	Auth4cData::PermissionMap& permissionS( _prole->permissions() );
	for( Auth4cData::PermissionMap::iterator iter = permissionS.begin();
			iter != permissionS.end();
			++iter )
	{
		new RolePermissionTreeWidgetItem( this, &iter->second );
	}
}

void RoleTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// RolePermissionTreeWidgetItem
//

RolePermissionTreeWidgetItem::RolePermissionTreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::Permission* ppermission, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, ppermission, state ),
	_ppermission( ppermission )
{
	refresh();
}

RolePermissionTreeWidgetItem::~RolePermissionTreeWidgetItem()
{
}

void RolePermissionTreeWidgetItem::refresh()
{
	this->setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/perm_16.xpm" ) );
	setText( 0, QString::fromStdString( _ppermission->description() ) );
}

void RolePermissionTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// CustomerTreeWidgetItem
//

CustomerTreeWidgetItem::CustomerTreeWidgetItem( QTreeWidget* pparent, Auth4cData::Customer* pcustomer, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pcustomer, state ),
	_pcustomer( pcustomer )
{
	refresh();
}

CustomerTreeWidgetItem::~CustomerTreeWidgetItem()
{
}

void CustomerTreeWidgetItem::refresh()
{
	std::string strLegend;
	strLegend = _pcustomer->name();
	strLegend += " (";
	strLegend += _pcustomer->customer();
	strLegend += ")";
	setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/airline_16.xpm" ) );
	setText( 0, QString::fromStdString( strLegend ) );
}

void CustomerTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// DomainTreeWidgetItem
//

DomainTreeWidgetItem::DomainTreeWidgetItem( QTreeWidget* pparent, Auth4cData::Domain* pdomain, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pdomain, state ),
	_pdomain( pdomain )
{
	refresh();
}

DomainTreeWidgetItem::~DomainTreeWidgetItem()
{
}

void DomainTreeWidgetItem::refresh()
{
	std::string strLegend;
	strLegend = _pdomain->domain();
	strLegend += " (";
	strLegend += _pdomain->customer().customer();
	strLegend += ")";
	setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/domain_16.xpm" ) );
	setText( 0, QString::fromStdString( strLegend ) );

	for( Auth4cData::UserMap::iterator iter = _pdomain->users().begin();
			iter != _pdomain->users().end();
			++iter )
	{
		new UserTreeWidgetItem( this, &iter->second );
	}
}

void DomainTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// TeamTreeWidgetItem
//

TeamTreeWidgetItem::TeamTreeWidgetItem( QTreeWidget* pparent, Auth4cData::Team* pteam, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pteam, state ),
	_pteam( pteam )
{
	setChildIndicatorPolicy( QTreeWidgetItem::ShowIndicator );
	refresh();
}

TeamTreeWidgetItem::TeamTreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::Team* pteam, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pteam, state ),
	_pteam( pteam )
{
	refresh();
}

TeamTreeWidgetItem::~TeamTreeWidgetItem()
{
}

Auth4cData::Team* TeamTreeWidgetItem::team()
{
	return _pteam;
}

void TeamTreeWidgetItem::refresh()
{
	QString strLegend;
	strLegend = QString::fromStdString( _pteam->name() );
	setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/group_16.xpm" ) );
	setText( 0, strLegend );

	takeChildren();

	for( Auth4cData::TeamMemberMap::const_iterator iterMember = _pteam->members().begin();
			iterMember != _pteam->members().end();
			++iterMember )
	{
		new TeamMemberTreeWidgetItem( this, iterMember->second, Unchecked );
	}
}

void TeamTreeWidgetItem::doMenu()
{
	QMenu menu;

	QAction* pactionRefresh = menu.addAction( QObject::tr( "Re&fresh" ) );
	QAction* pactionReload = menu.addAction( QObject::tr( "Re&load" ) );
	QAction* pactionExport = menu.addAction( QObject::tr( "E&xport" ) );

	QAction* pactionSelected = menu.exec( QCursor::pos() );

	if( pactionSelected == pactionRefresh )
	{
		treeWidget()->setCursor( Qt::BusyCursor );

		try
		{
			Auth4cDataXML::Factory::team( *_pteam, false );
			refresh();
		}
		catch( ... )
		{
		}

		treeWidget()->unsetCursor();
	}
	else if( pactionSelected == pactionReload )
	{
		treeWidget()->setCursor( Qt::BusyCursor );

		try
		{
			Auth4cDataXML::Factory::team( *_pteam, true );
			refresh();
		}
		catch( ... )
		{
		}

		treeWidget()->unsetCursor();
	}
	else if( pactionSelected == pactionExport )
	{
		treeWidget()->setCursor( Qt::BusyCursor );

		try
		{
			Auth4cData::Exporter* pexporter = Auth4cData::Factory::exporter();
			if( NULL != pexporter )
			{
				pexporter->exportFile( *_pteam );
			}
		}
		catch( ... )
		{
		}

		treeWidget()->unsetCursor();
	}
}

void TeamTreeWidgetItem::expanded()
{
	if( !_pteam->members().size() )
	{
		treeWidget()->setCursor( Qt::BusyCursor );

		try
		{
			Auth4cDataXML::Factory::team( *_pteam, false );
			refresh();
		}
		catch( ... )
		{
		}

		treeWidget()->unsetCursor();
	}
}


//////////////////////////////////////////////////////////////////////////
//
// TeamMemberRootTreeWidgetItem
//

TeamMemberRootTreeWidgetItem::TeamMemberRootTreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::Team* pteam, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, pteam, state ),
	_pteam( pteam )
{
	refresh();
}

TeamMemberRootTreeWidgetItem::~TeamMemberRootTreeWidgetItem()
{
}

void TeamMemberRootTreeWidgetItem::refresh()
{
	setText( 0, QObject::tr( "Members" ) );

	takeChildren();

	for( Auth4cData::TeamMemberMap::const_iterator iterMember = _pteam->members().begin();
			iterMember != _pteam->members().end();
			++iterMember )
	{
		new TeamMemberTreeWidgetItem( this, iterMember->second );
	}
}

void TeamMemberRootTreeWidgetItem::doMenu()
{
	QMenu menu;

	QAction* pactionRefresh = menu.addAction( QObject::tr( "Re&fresh" ) );
	QAction* pactionReload = menu.addAction( QObject::tr( "Re&load" ) );

	QAction* pactionSelected = menu.exec( QCursor::pos() );

	if( pactionSelected == pactionRefresh )
	{
		treeWidget()->setCursor( Qt::BusyCursor );

		try
		{
			Auth4cDataXML::Factory::team( *_pteam, false );
			refresh();
		}
		catch( ... )
		{
		}

		treeWidget()->unsetCursor();
	}
	else if( pactionSelected == pactionReload )
	{
		treeWidget()->setCursor( Qt::BusyCursor );

		try
		{
			Auth4cDataXML::Factory::team( *_pteam, true );
			refresh();
		}
		catch( ... )
		{
		}

		treeWidget()->unsetCursor();
	}
}


//////////////////////////////////////////////////////////////////////////
//
// UserTreeWidgetItem
//

UserTreeWidgetItem::UserTreeWidgetItem( TreeWidgetItem* pparent, Auth4cData::User* puser, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, puser, state ),
	_puser( puser )
{
	refresh();
}

UserTreeWidgetItem::~UserTreeWidgetItem()
{
}

Auth4cData::User* UserTreeWidgetItem::user()
{
	return _puser;
}

void UserTreeWidgetItem::refresh()
{
	setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/user_16.xpm" ) );
	setText( 0, QString::fromStdString( _puser->userId() ) );
}

void UserTreeWidgetItem::doMenu()
{
}


//////////////////////////////////////////////////////////////////////////
//
// TeamMemberTreeWidgetItem
//

TeamMemberTreeWidgetItem::TeamMemberTreeWidgetItem( TreeWidgetItem* pparent, const Auth4cData::TeamMember& member, CheckState state /*= Uncheckable*/ )
	: TreeWidgetItem( pparent, &_member, state ),
	_member( member )
{
	refresh();
}

TeamMemberTreeWidgetItem::~TeamMemberTreeWidgetItem()
{
}

Auth4cData::TeamMember* TeamMemberTreeWidgetItem::member()
{
	return &_member;
}

void TeamMemberTreeWidgetItem::refresh()
{
	QString strLegend;
	strLegend = QString::fromUtf8( _member.name().c_str() );
	strLegend += " (";
	strLegend += QString::fromStdString( _member.domain().domain() );
	strLegend += "/";
	strLegend += QString::fromStdString( _member.userId() );
	strLegend += ")";
	if( _member.isLead() )
	{
		setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/lead_16.xpm" ) );
	}
	else
	{
		setIcon( 0, QIcon( ":/Auth4cUIComponents/Resources/user_16.xpm" ) );
	}
	setText( 0, strLegend );
}

void TeamMemberTreeWidgetItem::doMenu()
{
}
