/*

  $Id: Auth4cDApp.cpp 699 2018-12-11 17:51:05Z john $

  Copyright (c) 2008, The 4c Company. All rights reserved.
*/

#define _WINSOCKAPI_

#include <auth4cd/Auth4cDApp.h>
#include <Auth4cDataXML/Util.h>
#include <Auth4cDataDB/Auth4cFactory.h>

#include <Base4c/Exception.h>
#include <Base4c/EventLogger.h>

#include <librfcdal/DalDatabaseException.h>

#include <time.h>

#include <string>
#include <fstream>

#if defined (LINUX)
#	include <fcntl.h>
#	include "netinet/in.h"
#	include "arpa/inet.h"
#	include "netdb.h"
#	include <sys/socket.h>
#	include <sys/ioctl.h>
#	include <sys/stat.h>
#	include <signal.h>
#	include <sys/types.h>
#	include <unistd.h>
#elif defined (_WINDOWS)
#	include <process.h>
#	include <Ws2tcpip.h>
#endif /* LINUX */

#include <time.h>
//#include <iostream>
#include <librfc/XMLDoc.hpp>

using namespace Auth4cD;


Auth4cDApp::Auth4cDApp()
	: _bIsTerminating( false ),
	_bFulfillmentThreadTerminated( false ),
	_bListeningThreadTerminated( false ),
	_s( -1 ),
	_nPort( 777 ),
	_strPIDFile( "/var/run/auth4cd.pid" ),
	_strRC( "/etc/auth4cd/rcauth4cd" ),
#if defined (LINUX)
	_strKeyDir("/etc/auth4cd/"),
#elif defined (_WINDOWS)
	_strKeyDir(""),
#endif /* LINUX */
	_bIdle( true )
{
}

Auth4cDApp theApp;

Auth4cDApp::~Auth4cDApp()
{
}

int Auth4cDApp::initInstance( int nArgc, char** pszArgvS )
{
	int nRC = 0;
	bool bHelp( false );
//	bool bDebug( false );
	bool bGenerate( false );
	bool bDaemon( false );
	std::string strEntropy("");

	for( int nI = 1; nI <nArgc; ++nI )
	{
		if( pszArgvS[nI][0] == '-' )
		{
			switch( pszArgvS[nI][1] )
			{
				case 'd':
					bDaemon=true;
					break;

				case 'D':
					bDaemon = false;
					//bDebug = true;
					break;

				case 'p':
					_strPIDFile = &pszArgvS[nI][2];
					break;

				case 'l':
					if( strlen( pszArgvS[nI] ) > 2 )
					{
						int nPort = atoi( &pszArgvS[nI][2] );
						if( nPort )
							_nPort = nPort;
					}
					break;

				case 'k':
					_strKeyDir = pszArgvS[nI][2];
					if( ( _strKeyDir[ ( _strKeyDir.length() - 1 ) ] != '/' ) &&
						( _strKeyDir[ ( _strKeyDir.length() - 1 ) ] != '\\' ) )
					{
						_strKeyDir += "/";
					}
					break;

				case 'g':
					strEntropy = pszArgvS[nI][2];
					bGenerate = true;
					break;

				case '?':
				default:
					bHelp = true;
			}
		}
		else
			_strRC = pszArgvS[nI];
	}

	if( bGenerate )
	{
		return generateKeyPair( strEntropy );
	}

	if( bHelp )
	{
		return nHelp;
	}

	try
	{
		std::ifstream conIn( _strRC.c_str() );
	
		if ( !conIn.is_open() )
		{
			std::string strLegend( "Failed to read from file: " );
			strLegend += _strRC;
			Base4c::EventLogger::theLogger()->logError( strLegend, Base4c::EventLogger::daemon );
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_CONNECTOPENFAILED, _strRC );
		}
		
		std::string strConnect;
		conIn >> strConnect;

		Auth4cDataDB::Factory::initialize( strConnect, Auth4cDataDB::Factory::modeAuthenticator );
	}
	catch( const Auth4cData::Exception& e )
	{
		Base4c::EventLogger::theLogger()->logError( e.legend(), Base4c::EventLogger::daemon );
	}
	catch( const Base4c::Exception& e )		
	{
		Base4c::EventLogger::theLogger()->logError( e.legend(), Base4c::EventLogger::daemon );
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		Base4c::EventLogger::theLogger()->logError( e.nativeExplanation(), Base4c::EventLogger::daemon );
	}

	union prng_state prng;

	yarrow_start( &prng );
	yarrow_add_entropy( (unsigned char*)"hello", 5, &prng );
	yarrow_ready( &prng );

	/* make a 1024-bit RSA key */
	if( CRYPT_ERROR == rsa_make_key( &prng, find_prng( (char*)"yarrow" ), 1024/8, 65537, &_key ) )
	{
		return -1;
	}

	_s = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	if( _s == -1 )
	{
		Base4c::EventLogger::theLogger()->logError( "Failed to create socket", Base4c::EventLogger::daemon );
		throw Base4c::Exception( __FILE__, __LINE__, Base4c::B4CX_SEELEGEND, "Failed to create socket" );
	}

	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_port = htons( _nPort );
	addr.sin_addr.s_addr = INADDR_ANY;
	if( bind( _s, (const struct sockaddr*)&addr, sizeof( addr ) ) == -1 )
	{
#if defined (LINUX)
		close( _s );
#elif defined (_WINDOWS)
		closesocket( _s );
#endif /* LINUX */

		Base4c::EventLogger::theLogger()->logError( "Failed to bind to socket", Base4c::EventLogger::daemon );
		throw Base4c::Exception( __FILE__, __LINE__, Base4c::B4CX_SEELEGEND, "Failed to bind to socket" );
	}

	if( listen( _s, SOMAXCONN ) == -1 )
	{
#if defined (LINUX)
		close( _s );
#elif defined (_WINDOWS)
		closesocket( _s );
#endif /* LINUX */
		Base4c::EventLogger::theLogger()->logError( "Failed to listen on socket", Base4c::EventLogger::daemon );
		throw Base4c::Exception( __FILE__, __LINE__, Base4c::B4CX_SEELEGEND, "Failed to listen on socket" );
	}

	_socketS.add( ClientSocket::clientListen, _s, std::string() );

	if( bDaemon )
	{
#if defined (LINUX)
//		daemon( NULL, NULL ); // This is the old way...
		daemon( 1, 0 );
#endif /* LINUX */
	}

	// have to get pid after daemon is running as daemon() changes pid
	if( _strPIDFile.length() )
	{
#if defined (LINUX)
		int pid = getpid();

		int hPID = open( _strPIDFile.c_str(), O_CREAT | O_TRUNC | O_WRONLY,
						 S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH );
		if( hPID != -1 )
		{
			char szPID[16];
			sprintf( szPID, "%i\n", pid );
			write( hPID, szPID, strlen( szPID ) );
			close( hPID );
		}
		else
			_strPIDFile = "";
#endif /* LINUX */
	}

	Base4c::EventLogger::theLogger()->logDebug( "Running...", Base4c::EventLogger::daemon );

	return nRC;
}

int Auth4cDApp::run( int /*nStartCode*/ )
{
	Base4c::EventLogger::theLogger()->logMessage( "Auth4c Daemon starting...", Base4c::EventLogger::daemon );

#if defined (_WINDOWS)
	_beginthread( listeningThread, 2048, this );
#elif defined (LINUX)
	pthread_t thread;
	pthread_create( &thread, NULL, listeningThread, this );
	pthread_detach( thread );
#endif /* _WINDOWS */

		fulfillmentThread( this );

		int nMaxWait = 5;
		while( _socketS.count() && --nMaxWait )
#if defined (LINUX)
			sleep( 1 );
#elif defined (_WINDOWS)
			Sleep( 1000 );
#endif /* LINUX */

	return 0;
}

int Auth4cDApp::exitInstance( int nEndCode )
{
	if( nEndCode == nHelp )
	{
		std::cout << "Usage: auth4cd [-D] [-l<port>] [-p<pid_file>] [-I|-U] <config_file>\n";
		std::cout << "             -D don't run as daemon and output debug info\n";
		std::cout << "             -g<entropy> - specify entropy for new RSA key generation\n";
		std::cout << "             -k<directory> - specify directory for rcauth4cd and key files. Default /etc/auth4cd\n";
		std::cout << "             -l<port> - specify port to listen on. Default 5055\n";
		std::cout << "             -p<pid_file> defaults to /var/run/auth4cd.pid\n";
		std::cout << "             <config_file> defaults to /etc/rcauth4cd\n";
		std::cout << std::endl;
	}
	else
	{
		if( _strPIDFile.length() )
			unlink( _strPIDFile.c_str() );
	}

	return nEndCode;
}

rsa_key& Auth4cDApp::key()
{
	return _key;
}

bool Auth4cDApp::isTerminating() const
{
	return _bIsTerminating;
}

ClientSocketList& Auth4cDApp::sockets()
{
	return _socketS;
}

int Auth4cDApp::acceptClient( ClientSocket* pclient )
{
	int nS = 0;
	char szHost[128];
	char szService[16];

	struct sockaddr addr;
#if defined (LINUX)
	socklen_t nLen( sizeof addr );
#elif defined (_WINDOWS)
	int nLen( sizeof addr );
#endif /* LINUX */

	nS = accept( pclient->s(), &addr, &nLen );

	getnameinfo( &addr, nLen, szHost, sizeof szHost, szService, sizeof szService, 0 );

	if( nS != -1 )
	{
		_socketS.add( ClientSocket::clientConnect, nS, szHost );
		if( Base4c::EventLogger::theLogger()->logDebug( "Accepted client.", Base4c::EventLogger::daemon ) )
			std::cerr << "Accepted client " << nS << "\n";
	}

	return nS;
}

void shutdown( int nSignal )
{
	char szMsg[64];
	sprintf( szMsg, "Shutting down on signal %s", signame( nSignal ) ); 
	Base4c::EventLogger::theLogger()->logMessage( szMsg, Base4c::EventLogger::daemon );
	theApp.shutdown();
}

void ignore( int nSignal )
{
	char szMsg[64];
	sprintf( szMsg, "Ignoring signal %s", signame( nSignal ) ); 
	Base4c::EventLogger::theLogger()->logMessage( szMsg, Base4c::EventLogger::daemon );
}

const char* signame( int nSignal )
{
	const char* pszRtn = "unknown";

#if defined (LINUX)
	switch( nSignal )
	{
		case SIGHUP:
			pszRtn = "SIGHUP";
			break;
		case SIGINT:
			pszRtn = "SIGINT";
			break;
		case SIGQUIT:
			pszRtn = "SIGQUIT";
			break;
		case SIGILL:
			pszRtn = "SIGILL";
			break;
		case SIGTRAP:
			pszRtn = "SIGTRAP";
			break;
		case SIGABRT:	// SIGIOT
			pszRtn = "SIGABRT/SIGIOT";
			break;
		case SIGBUS:
			pszRtn = "SIGBUS";
			break;
		case SIGFPE:
			pszRtn = "SIGFPE";
			break;
		case SIGKILL:
			pszRtn = "SIGKILL";
			break;
		case SIGUSR1:
			pszRtn = "SIGUSR1";
			break;
		case SIGSEGV:
			pszRtn = "SIGSEGV";
			break;
		case SIGUSR2:
			pszRtn = "SIGUSR2";
			break;
		case SIGPIPE:
			pszRtn = "SIGPIPE";
			break;
		case SIGALRM:
			pszRtn = "SIGALRM";
			break;
		case SIGTERM:
			pszRtn = "SIGTERM";
			break;
		case SIGSTKFLT:
			pszRtn = "SIGSTKFLT";
			break;
		case SIGCLD: // SIGCHLD
			pszRtn = "SIGCLD/SIGCHLD";
			break;
		case SIGCONT:
			pszRtn = "SIGCONT";
			break;
		case SIGSTOP:
			pszRtn = "SIGSTOP";
			break;
		case SIGTSTP:
			pszRtn = "SIGTSTP";
			break;
		case SIGTTIN:
			pszRtn = "SIGTTIN";
			break;
		case SIGTTOU:
			pszRtn = "SIGTTOU";
			break;
		case SIGURG:
			pszRtn = "SIGURG";
			break;
		case SIGXCPU:
			pszRtn = "SIGXCPU";
			break;
		case SIGXFSZ:
			pszRtn = "SIGXFSZ";
			break;
		case SIGVTALRM:
			pszRtn = "SIGVTALRM";
			break;
		case SIGPROF:
			pszRtn = "SIGPROF";
			break;
		case SIGWINCH:
			pszRtn = "SIGWINCH";
			break;
		case SIGPOLL: // SIGIO
			pszRtn = "SIGPOLL/SIGIO";
			break;
		case SIGPWR:
			pszRtn = "SIGPWR";
			break;
		case SIGSYS:
			pszRtn = "SIGSYS";
			break;
	}
#endif /* LINUX */

	return pszRtn;
}

void Auth4cDApp::shutdown()
{
	_bIsTerminating = true;
}

#if defined (_WINDOWS)
void Auth4cDApp::fulfillmentThread( void* pvContext )
#elif defined (LINUX)
void* Auth4cDApp::fulfillmentThread( void* pvContext )
#endif /* _WINDOWS */
{
	Auth4cDApp* This = (Auth4cDApp*)pvContext;

	Base4c::EventLogger::theLogger()->logDebug( "Fulfillment thread started", Base4c::EventLogger::daemon );

	while( !This->isTerminating() )
	{
#if defined (LINUX)
		sleep( 1 );
#elif defined (_WINDOWS)
		Sleep( 1000 );
#endif /* LINUX */
	}

	This->_bFulfillmentThreadTerminated = true;

#if defined (LINUX)
	return NULL;
#endif /* LINUX */
}

#if defined (_WINDOWS)
void Auth4cDApp::listeningThread( void* pvContext )
#elif defined (LINUX)
void* Auth4cDApp::listeningThread( void* pvContext )
#endif /* _WINDOWS */
{
	Auth4cDApp* This = (Auth4cDApp*)pvContext;

	fd_set set;
	int nMax = 0;
	ClientSocket* pskt;
	int rc;
#if defined (LINUX)
	struct timespec ts;
	sigset_t sigs;

	sigfillset( &sigs );
#elif defined (_WINDOWS)
	struct timeval tv;
#endif /* LINUX */


	try
	{
		while( !This->_bIsTerminating )
		{
			FD_ZERO( &set );
			pskt = This->_socketS.next( NULL );
			while( pskt != NULL )
			{
				if( pskt->type() == ClientSocket::clientListen )
				{
					FD_SET( pskt->s(), &set );
					if( pskt->s() > nMax )
						nMax = pskt->s();
				}
				pskt = This->_socketS.next( pskt );
			}

			nMax++;

//		tv needs to be reset every time as under Linux, when select returns, tv
//		has been modified to hold the outstanding time of the select call for
//		its timeout.
#if defined (LINUX)
			ts.tv_sec = 1;
			ts.tv_nsec = 0;

			rc = pselect( nMax, &set, NULL, NULL, &ts, &sigs );
#elif defined (_WINDOWS)
			tv.tv_sec = 0;
			tv.tv_usec = 100;

			rc = select( nMax, &set, NULL, NULL, &tv );
#endif /* LINUX */

			if( rc )
			{
				pskt = This->_socketS.next( NULL );
				while( pskt != NULL )
				{
					if( FD_ISSET( pskt->s(), &set ) )
					{
						This->acceptClient( pskt );
					}
					pskt = This->_socketS.next( pskt );
				}
			}

		}

		This->_socketS.remove( This->_s );
	}
	catch( ... )
	{
		Base4c::EventLogger::theLogger()->logUnknown( Base4c::EventLogger::daemon );
	}

	This->_bListeningThreadTerminated = true;
	
#if defined (_WINDOWS)
	_endthread();
#elif defined (LINUX)
	pthread_exit( NULL );

	return NULL;
#endif /* _WINDOWS */
}

int Auth4cDApp::generateKeyPair( const std::string& strEntropy )
{
	unsigned char in[4096]; //, out[4096];
	int x; //, y;
	struct rsa_key key;
	union prng_state prng;

	/* start prng */
	yarrow_start(&prng);
	yarrow_add_entropy( (unsigned char*)strEntropy.c_str(), strEntropy.length(), &prng);
	yarrow_ready(&prng);
	
	for (x = 0; x < 8; x++) in[x] = x;
	sprintf( (char*)in, "bugsythecat" );

	/* make a 1024-bit RSA key */
	const char* pszYarrow = "yarrow";
	if (rsa_make_key(&prng, find_prng( (char*)pszYarrow ), 1024/8, 65537, &key) == CRYPT_ERROR) { printf("Error: %s\n", crypt_error); return -1; }

	std::ofstream pubOut( ( _strKeyDir + "nakey.pub" ).c_str() );
	std::ofstream prvOut( ( _strKeyDir + "nakey.prv" ).c_str() );

	pubOut << key.e.sign << "\n" << key.e.alloc << "\n" << key.e.used << "\n" << key.e.dp;
	for( unsigned int i = 0; i < key.e.alloc; ++i )
	{
		pubOut << key.e.dp[i];
	}
	prvOut << key.d.sign << "\n" << key.d.alloc << "\n" << key.d.used << "\n" << key.d.dp;
	for( unsigned int i = 0; i < key.d.alloc; ++i )
	{
		prvOut << key.d.dp[i];
	}

	return 0;
}

struct rsa_key Auth4cDApp::keyStatic()
{
	return _keyStatic;
}
