/*

  $Id: ClientSocket.cpp 739 2020-09-30 15:36:00Z john $

  Copyright (c) 2008, The 4c Company. All rights reserved.
*/

#include <auth4cd/ClientSocket.h>

#include <auth4cd/Auth4cDApp.h>

#include <Auth4cData/A4CAuth4c.h>
#include <Auth4cData/Factory.h>
#include <Auth4cData/PasswordAuthenticationToken.h>
#include <Auth4cData/Exception.h>
#include <Auth4cDataXML/Auth4cFactory.h>
#include <Auth4cDataDB/Auth4cFactory.h>

#include <Base4c/EventLogger.h>
#include <Base4c/Exception.h>

#if defined (_UNIX_) || defined (LINUX)
#	include <sys/types.h>
#	include <unistd.h>
#	include <arpa/inet.h>
#	include <netdb.h>
#	include <sys/ioctl.h>
#	include <signal.h>
#elif defined (_WINDOWS)
#	include <process.h>
#else
#	error No platform defined !
#endif

#include <librfc/XMLDoc.hpp>
#include <librfc/XMLNode.hpp>

#include <sstream>

#define MTU 2048

using namespace Auth4cD;

Auth4cData::Product g_product( A4CAuth4C::nA4CD_Auth4c );

//////////////////////////////////////////////////////////////////////////
//
// ClientSocketList
//

ClientSocketList::ClientSocketList()
	: _pfirst( NULL ),
	_plast( NULL ),
	_mx( "ClientSocketList::mmx" ),
	_nCount( 0 )
{
}

ClientSocketList::~ClientSocketList()
{
	ClientSocket* pcurrent = _pfirst;
	ClientSocket* pnext;

	while( pcurrent != NULL )
	{
		pnext = pcurrent->_pnext;
		delete pcurrent;
		pcurrent = pnext;
	}

	_pfirst = _plast = NULL;
}

void ClientSocketList::add( ClientSocket::Type type, int s, const std::string& strHost )
{
	_mx.lock( "ClientSocketList::add" );

	try
	{
		ClientSocket* padd = new ClientSocket( type, s, strHost );

		if( _pfirst == NULL )
		{
			_pfirst = _plast = padd;
		}
		else
		{
			padd->_pprev = _plast;
			_plast = _plast->_pnext = padd;
		}

		++_nCount;
	}
	catch( ... )
	{
	}

	_mx.unlock();
}

void ClientSocketList::remove( int s )
{
	_mx.lock( "ClientSocketList::remove" );

	try
	{
		ClientSocket* premove = find( s );

		if( premove == NULL )
		{
			_mx.unlock();
			return;
		}

		--_nCount;

		if( premove->_pprev == NULL )
		{
			_pfirst = premove->_pnext;
		}
		else
		{
			premove->_pprev->_pnext = premove->_pnext;
		}

		if( premove->_pnext == NULL )
		{
			_plast = premove->_pprev;
		}
		else
		{
			premove->_pnext->_pprev = premove->_pprev;
		}

		delete premove;
	}
	catch( ... )
	{
	}

	_mx.unlock();
}

ClientSocket* ClientSocketList::find( int s )
{
	ClientSocket* pcurrent = _pfirst;
	ClientSocket* pfound = NULL;

	_mx.lock( "ClientSocketList::find" );

	try
	{
		while( ( pcurrent != NULL ) && ( pfound == NULL ) )
		{
			if( pcurrent->_s == s )
				pfound = pcurrent;
			else
				pcurrent = pcurrent->_pnext;
		}
	}
	catch( ... )
	{
	}

	_mx.unlock();

	return pfound;
}

ClientSocket* ClientSocketList::next( ClientSocket* pcurrent )
{
	if( pcurrent == NULL )
		return _pfirst;
	else
		return pcurrent->_pnext;
}

void ClientSocketList::broadcastMessage( const std::string& strMessage )
{
	_mx.lock( "ClientSocketList::broadcastMessage" );

	try
	{
		ClientSocket* pskt = NULL;
		pskt = next( pskt );
		while( pskt != NULL )
		{
			if( pskt->type() == ClientSocket::clientConnect )
				pskt->postMessage( strMessage );
			pskt = next( pskt );
		}
	}
	catch( ... )
	{
	}

	_mx.unlock();
}

void ClientSocketList::lock()
{
	_mx.lock( "ClientSocketList::lock" );
}

void ClientSocketList::unlock()
{
	_mx.unlock();
}

//////////////////////////////////////////////////////////////////////////
//
// ClientSocket
//

ClientSocket::ClientSocket( Type type, int s, const std::string& strHost )
	: _pnext( NULL ),
	_pprev( NULL ),
	_type( type ),
	_s( s ),
	_strHost( strHost )
{
	if( _type == clientConnect )
	{
#if defined (LINUX)
		pthread_t thread;
		pthread_create( &thread, NULL, clientProc, this );
		pthread_detach( thread );
#elif defined (_WINDOWS)
		_beginthread(clientProc, 1024, this );
#endif /* LINUX */
	}
}

ClientSocket::~ClientSocket()
{
#if defined (LINUX)
	close( _s );
#elif defined (_WINDOWS)
	closesocket( _s );
#endif /* LINUX */
}

const Auth4cData::AuthenticatedUser& ClientSocket::user() const
{
	return _user;
}

ClientSocket::Type ClientSocket::type() const
{
	return _type;
}

int ClientSocket::s() const
{
	return _s;
}

#if defined (LINUX)
void* ClientSocket::clientProc( void* pvContext )
#elif defined (_WINDOWS)
void ClientSocket::clientProc( void* pvContext )
#endif
{
	ClientSocket* This = (ClientSocket*)pvContext;

	This->_bContinue = true;

	fd_set set;
	int rc;

#if defined (LINUX)
	struct timespec ts;
	sigset_t sigs;

	sigfillset( &sigs );
#elif defined (_WINDOWS)
	struct timeval tv;
#endif /* LINUX */

	try
	{
		while( !theApp.isTerminating() && This->_bContinue )
		{
			FD_ZERO( &set );
			FD_SET( This->s(), &set );
			int nMax = This->s() + 1;

//		tv needs to be reset every time as under Linux, when select returns, tv
//		has been modified to hold the outstanding time of the select call for
//		its timeout.
#if defined (LINUX)
			ts.tv_sec = 1;
			ts.tv_nsec = 0;

			rc = pselect( nMax, &set, NULL, NULL, &ts, &sigs );
#elif defined (_WINDOWS)
			tv.tv_sec = 0;
			tv.tv_usec = 100;

			rc = select( nMax, &set, NULL, NULL, &tv );
#endif /* LINUX */

			if( rc )
			{
				This->readClient();
			}
		}
		theApp.sockets().remove( This->s() );
	}
	catch( const Rfc::XMLException& e )
	{
		theApp.sockets().remove( This->s() );
		std::ostringstream os;
		os << "In ClientSocket::clientProc XMLException: " << e.reason();
		Base4c::EventLogger::theLogger()->logError( os.str(), Base4c::EventLogger::daemon );
	}
	catch( const Base4c::Exception& e )
	{
		theApp.sockets().remove( This->s() );
		std::ostringstream os;
		os << "In ClientSocket::clientProc Base4c::Exception: " << e; //.explainErrorCode( e.errorCode() ) << " : " << e.legend();
		Base4c::EventLogger::theLogger()->logError( os.str(), Base4c::EventLogger::daemon );
	}
	catch( const Auth4cData::Exception& e )
	{
		theApp.sockets().remove( This->s() );
		std::ostringstream os;
		os << "In ClientSocket::clientProc Auth4cData::Exception: " << e; //.explainErrorCode( e.errorCode() ) << " : " << e.legend();
		Base4c::EventLogger::theLogger()->logError( os.str(), Base4c::EventLogger::daemon );
	}
	catch( const RfcDal::DalDatabaseException& e )
	{
		theApp.sockets().remove( This->s() );
		std::ostringstream os;
		os << "In ClientSocket::clientProc RfcDal::DalDatabaseException: " << e; //.explainErrorCode( e.errorCode() ) << " : " << e.legend();
		Base4c::EventLogger::theLogger()->logError( os.str(), Base4c::EventLogger::daemon );
	}
	catch( const std::exception& e )
	{
		theApp.sockets().remove( This->s() );
		std::ostringstream os;
		os << "In ClientSocket::clientProc std::exception: " << e.what();
		Base4c::EventLogger::theLogger()->logError( os.str(), Base4c::EventLogger::daemon );
	}
	catch( ... )
	{
		theApp.sockets().remove( This->s() );
		Base4c::EventLogger::theLogger()->logError( "Unknown exception caught in ClientSocket::clientProc", Base4c::EventLogger::daemon );
	}

#if defined (LINUX)
	pthread_exit( NULL );

	return NULL;
#elif defined (_WINDOWS)
	_endthread();
#endif /* LINUX */
}

int ClientSocket::readClient()
{
	int nRC = 0;
	unsigned long ulLen = 0;

#if defined (LINUX)
	nRC = ioctl( _s, FIONREAD, &ulLen );
#elif defined (_WINDOWS)
	nRC = ioctlsocket( _s, FIONREAD, &ulLen );
#endif /* LINUX */

	if( ulLen == 0 )
	{
		_bContinue = false;
		Rfc::String strLegend;
		strLegend.format( "Closing client %i\n", _s );
		Base4c::EventLogger::theLogger()->logDebug( (const char*)strLegend, Base4c::EventLogger::daemon );
	}
	else
	{
		char* pszBuffer = new char[( ulLen + 1 )];
		pszBuffer[ulLen] = '\0';
		recv( _s, pszBuffer, ulLen, 0 );
		_strMessages += pszBuffer;
		delete[] pszBuffer;
		if( _strMessages.at( _strMessages.length() - 1 ) == '\x04' )
		{
			processClientRequests();
			_strMessages = "";
		}
	}

	return nRC;
}

void ClientSocket::processClientRequests()
{
	Rfc::String strMessages( _strMessages.c_str() );
	Rfc::String strMessage;
	int nField( 0 );

	strMessage = strMessages.field( '\x04', nField );
	while( strMessage.length() )
	{
		processClientRequest( strMessage );
		strMessage = strMessages.field( '\x04', ++nField );
	}
}

void ClientSocket::processClientRequest( const Rfc::String& strMessage )
{
	Rfc::String strRequest( "unknown" );
	long long nRequestId = -1;
	Rfc::String strRequestId;
	Rfc::String strRefId;

	Rfc::XMLDoc docR( "auth4cservice", "1.0" );
	Rfc::XMLNode* pnodeR = docR.rootNode();

	try
	{
		Rfc::XMLDoc docQ( (const char*)strMessage );
		Rfc::XMLNode* pnodeQ = docQ.rootNode();
		strRequest = pnodeQ->attribute( "request" );
		if( !strRequest.isNull() && strRequest.length() )
		{
			if( strRequest == "about" )
			{
				about( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "reloadfactory" )
			{
				reloadFactory( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "initialize" )
			{
				initialize( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "authenticate" )
			{
				authenticate( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "persistsession" )
			{
				persistSession( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "retrievesession" )
			{
				retrieveSession( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "killsession" )
			{
				killSession( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "validate" )
			{
				validate( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "authenticateduser" )
			{
				authenticatedUser( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "passwd" )
			{
				passwd( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "user" )
			{
				user( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "team" )
			{
				team( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "teams" )
			{
				teams( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "domains" )
			{
				domains( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "profiles" )
			{
				profiles( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "product" )
			{
				product( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "products" )
			{
				products( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "customers" )
			{
				customers( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else if( strRequest == "who" )
			{
				who( pnodeR, pnodeQ );
				postMessage( (const char*)docR.stream() );
			}
			else
			{
				unknownRequest( pnodeR, pnodeQ, strRequest );
				postMessage( (const char*)docR.stream() );
			}
		}
	}
	catch( const Rfc::XMLException& e )
	{
		std::ostringstream os;
		os << "XMLException: " << e.reason();
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, os.str() );
		Auth4cDataXML::Factory::Exception::toXML( pnodeR, e1 );
		pnodeR->attribute( "request", strRequest );
		strRequestId.format( "%i", nRequestId );
		pnodeR->attribute( "requestid", strRequestId );
		pnodeR->attribute( "response", "exception" );
		postMessage( (const char*)docR.stream() );
		return;														// EXIT POINT
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		Auth4cDataXML::Factory::Exception::toXML( pnodeR, e1 );
		pnodeR->attribute( "request", strRequest );
		strRequestId.format( "%i", nRequestId );
		pnodeR->attribute( "requestid", strRequestId );
		pnodeR->attribute( "response", "exception" );
		postMessage( (const char*)docR.stream() );
		return;														// EXIT POINT
	}
}

void ClientSocket::postMessage( const std::string& strMessage )
{
	std::string strLocal( strMessage );
	strLocal += "\x04";
	send( _s, strLocal.c_str(), strLocal.length(), 0 );
}

void ClientSocket::about( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	pnodeResponse->attribute( "response", "about" );
	Rfc::XMLNode* pnodeAbout = pnodeResponse->addNode( "about", Rfc::String() );
	pnodeAbout->addNode( "Auth4cData", "Factory", Auth4cData::Factory::version().text().c_str() );
	pnodeAbout->addNode( "Auth4cDataDB", "Factory", Auth4cDataDB::Factory::version().text().c_str() );
	pnodeAbout->addNode( "Auth4cDataXML", "Factory", Auth4cDataXML::Factory::version().text().c_str() );

#if defined (DEBUG)
	std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
}

void ClientSocket::reloadFactory( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		Auth4cData::Product productAuth4c( A4CAuth4C::nA4CD_Auth4c );
		Auth4cData::Role roleAdministrator( productAuth4c, A4CAuth4C::nA4CR_Administrator );
		if( !_user.hasRole( roleAdministrator ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_PERMISSIONDENIED, "Administrator", true );
		}

		Auth4cDataDB::Factory::reinitialize();

		pnodeResponse->attribute( "response", "reloadfactory" );
		pnodeResponse->addNode( "reinit", "ok" );
	}
	catch( const Auth4cData::Exception& e )
	{
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::initialize( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		Rfc::String strVersion = pnodeRequest->attribute( "version" );
		if( strVersion.isNull() )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_VERSIONMISMATCH, "No client version specified.", true );
		}

		Base4c::Version versionClient( Base4c::Version::fromText( (const char*)strVersion ) );
		// compare client version with server version
		if( 0 )
		{
			std::ostringstream os;
			os << "Client version " << versionClient.text() << " is incompatible with server version " << Auth4cData::Factory::version().text() << ".";
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_VERSIONMISMATCH, os.str(), true );
		}

		unsigned char szEncode[4096];
		int nLen;

		const rsa_key& key( theApp.key() );
		Rfc::String strVal;

		pnodeResponse->attribute( "response", "rsa" );
		Rfc::XMLNode* pnodeRSA = pnodeResponse->addNode( "rsa", Rfc::String() );

		Rfc::XMLNode* pnodeE = pnodeRSA->addNode( "e", Rfc::String() );
		strVal.format( "%i", key.e.sign );
		pnodeE->addNode( "sign", strVal );
		strVal.format( "%u", key.e.alloc );
		pnodeE->addNode( "alloc", strVal );
		strVal.format( "%u", key.e.used );
		pnodeE->addNode( "used", strVal );
		nLen = sizeof szEncode;
		base64_encode( (unsigned char*)key.e.dp, key.e.alloc * 2, szEncode, &nLen );
		szEncode[nLen] = '\0';
		pnodeE->addNode( "dp", (const char*)szEncode );

/*		We don't send the private key across the wire, except for debugging purposes */
/*		XMLNode* pnodeD = pnodeRSA->addNode( "d", String() );
		strVal.format( "%i", key.d.sign );
		pnodeD->addNode( "sign", strVal );
		strVal.format( "%u", key.d.alloc );
		pnodeD->addNode( "alloc", strVal );
		strVal.format( "%u", key.d.used );
		pnodeD->addNode( "used", strVal );
		nLen = sizeof szEncode;
		base64_encode( (unsigned char*)key.d.dp, key.d.alloc * 2, szEncode, &nLen );
		szEncode[nLen] = '\0';
		pnodeD->addNode( "dp", (const char*)szEncode );*/

		Rfc::XMLNode* pnodeN = pnodeRSA->addNode( "N", Rfc::String() );
		strVal.format( "%i", key.N.sign );
		pnodeN->addNode( "sign", strVal );
		strVal.format( "%u", key.N.alloc );
		pnodeN->addNode( "alloc", strVal );
		strVal.format( "%u", key.N.used );
		pnodeN->addNode( "used", strVal );
		nLen = sizeof szEncode;
		base64_encode( (unsigned char*)key.N.dp, key.N.alloc * 2, szEncode, &nLen );
		szEncode[nLen] = '\0';
		pnodeN->addNode( "dp", (const char*)szEncode );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::authenticate( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		unsigned char szDecode[4096];
		int nLenDecode = sizeof szDecode;
		unsigned char szDecrypt[4096];
		int nLenDecrypt = sizeof szDecrypt;
		unsigned char szToken[4096];
		int nRCDecode( 0 );

		const Rfc::XMLNode* pnodeToken = pnodeRequest->findChild( "authenticationtoken" );
		if( ( NULL == pnodeToken ) ||
			( pnodeToken->value().isNull() ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No authenticationtoken node in authenticate request." );
		}

		strcpy( (char*)szToken, (const char*)pnodeToken->value() );

		base64_decode( szToken, strlen( (const char*)szToken ), szDecode, &nLenDecode );

		if( ( nRCDecode = rsa_exptmod( szDecode, nLenDecode, szDecrypt, &nLenDecrypt, PK_PRIVATE, &theApp.key() ) ) == CRYPT_ERROR )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, crypt_error );
		}

		szDecrypt[nLenDecrypt] = '\0';

		unsigned char* pszDecrypt = szDecrypt;
		while( !*pszDecrypt && nLenDecrypt )
		{
			++pszDecrypt;
			--nLenDecrypt;
		}

		const Rfc::XMLNode* pnodeUser = pnodeRequest->findChild( "user" );
		if( NULL == pnodeUser )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No user node in authenticate request." );
		}

		Auth4cData::Profile profile;

		const Rfc::XMLNode* pnodeProfile = pnodeRequest->findChild( "profile" );
		if( NULL != pnodeProfile )
		{
			profile = Auth4cDataXML::Factory::Profile::fromXML( pnodeProfile );
		}

		Auth4cData::User user( Auth4cDataXML::Factory::User::fromXML( pnodeUser ) );

		try
		{
			Auth4cData::AuthenticatedUser userAuth( Auth4cDataDB::Factory::authenticate( user, Auth4cData::PasswordAuthenticationToken( (const char*)pszDecrypt ), profile ) );

			_user = userAuth;

			std::ostringstream os;
			os << "User " << user.domain().domain() << "/" << user.userId() << " successfully logged on from " << _strHost << ", using profile " << profile.profile();
			Base4c::EventLogger::theLogger()->logMessage( os.str(), Base4c::EventLogger::daemon );
		}
		catch( const Auth4cData::Exception& e )
		{
			std::ostringstream os;
			os << "User " << user.domain().domain() << "/" << user.userId() << " failed to log on from " << _strHost << ", using profile " << profile.profile() << " (" << e.legend() << ")";
			Base4c::EventLogger::theLogger()->logMessage( os.str(), Base4c::EventLogger::daemon );

			throw;
		}

		pnodeResponse->attribute( "response", "authenticateduser" );

//		Auth4cDataXML::Factory::AuthenticatedUser::toXML( pnodeResponse, userAuth );
		Auth4cDataXML::Factory::AuthenticatedUser::toXML( pnodeResponse, _user );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::validate( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		unsigned char szDecode[4096];
		int nLenDecode = sizeof szDecode;
		unsigned char szDecrypt[4096];
		int nLenDecrypt = sizeof szDecrypt;
		unsigned char szToken[4096];
		int nRCDecode( 0 );

		const Rfc::XMLNode* pnodeToken = pnodeRequest->findChild( "authenticationtoken" );
		if( ( NULL == pnodeToken ) ||
			( pnodeToken->value().isNull() ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No authenticationtoken node in authenticate request." );
		}

		strcpy( (char*)szToken, (const char*)pnodeToken->value() );

		base64_decode( szToken, strlen( (const char*)szToken ), szDecode, &nLenDecode );

		if( ( nRCDecode = rsa_exptmod( szDecode, nLenDecode, szDecrypt, &nLenDecrypt, PK_PRIVATE, &theApp.key() ) ) == CRYPT_ERROR )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, crypt_error );
		}

		szDecrypt[nLenDecrypt] = '\0';

		unsigned char* pszDecrypt = szDecrypt;
		while( !*pszDecrypt && nLenDecrypt )
		{
			++pszDecrypt;
			--nLenDecrypt;
		}

		const Rfc::XMLNode* pnodeUser = pnodeRequest->findChild( "user" );
		if( NULL == pnodeUser )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No user node in authenticate request." );
		}

		Auth4cData::Profile profile;

		const Rfc::XMLNode* pnodeProfile = pnodeRequest->findChild( "profile" );
		if( NULL != pnodeProfile )
		{
			profile = Auth4cDataXML::Factory::Profile::fromXML( pnodeProfile );
		}

		Auth4cData::User user( Auth4cDataXML::Factory::User::fromXML( pnodeUser ) );
		Auth4cData::AuthenticatedUser userAuth( Auth4cData::Factory::authenticate( user, Auth4cData::PasswordAuthenticationToken( (const char*)pszDecrypt ), profile ) );

		pnodeResponse->attribute( "response", "authenticateduser" );

		Auth4cDataXML::Factory::AuthenticatedUser::toXML( pnodeResponse, userAuth );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::authenticatedUser( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		if( !_user.session().str().length() )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_USERNOTFOUND, "There is no authenticated user.", true );
		}

		pnodeResponse->attribute( "response", "authenticateduser" );

		Auth4cDataXML::Factory::AuthenticatedUser::toXML( pnodeResponse, _user );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::passwd( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		unsigned char szDecode[4096];
		int nLenDecode = sizeof szDecode;
		unsigned char szDecrypt[4096];
		int nLenDecrypt = sizeof szDecrypt;
		unsigned char szToken[4096];
		int nRCDecode( 0 );

		const Rfc::XMLNode* pnodeTokenCurrent = pnodeRequest->findChild( "authenticationtokencurrent" );
		if( ( NULL == pnodeTokenCurrent ) ||
			( pnodeTokenCurrent->value().isNull() ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No authenticationtokencurrent node in passwd request." );
		}

		const Rfc::XMLNode* pnodeTokenNew = pnodeRequest->findChild( "authenticationtokennew" );
		if( ( NULL == pnodeTokenNew ) ||
			( pnodeTokenNew->value().isNull() ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No authenticationtokennew node in passwd request." );
		}

		strcpy( (char*)szToken, (const char*)pnodeTokenCurrent->value() );

		base64_decode( szToken, strlen( (const char*)szToken ), szDecode, &nLenDecode );

		if( ( nRCDecode = rsa_exptmod( szDecode, nLenDecode, szDecrypt, &nLenDecrypt, PK_PRIVATE, &theApp.key() ) ) == CRYPT_ERROR )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, crypt_error );
		}

		szDecrypt[nLenDecrypt] = '\0';

		unsigned char* pszDecrypt = szDecrypt;
		while( !*pszDecrypt && nLenDecrypt )
		{
			++pszDecrypt;
			--nLenDecrypt;
		}

		Auth4cData::PasswordAuthenticationToken authTokenCurrent( (const char*)pszDecrypt );

		strcpy( (char*)szToken, (const char*)pnodeTokenNew->value() );

		base64_decode( szToken, strlen( (const char*)szToken ), szDecode, &nLenDecode );

		if( ( nRCDecode = rsa_exptmod( szDecode, nLenDecode, szDecrypt, &nLenDecrypt, PK_PRIVATE, &theApp.key() ) ) == CRYPT_ERROR )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_AUTHENTICATIONFAILED, crypt_error );
		}

		szDecrypt[nLenDecrypt] = '\0';

		pszDecrypt = szDecrypt;
		while( !*pszDecrypt && nLenDecrypt )
		{
			++pszDecrypt;
			--nLenDecrypt;
		}

		Auth4cData::PasswordAuthenticationToken authTokenNew( (const char*)pszDecrypt );

		const Rfc::XMLNode* pnodeUser = pnodeRequest->findChild( "authenticateduser" );
		if( NULL == pnodeUser )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No authenticateduser node in passwd request." );
		}

		Auth4cData::Profile profile;

		Auth4cData::AuthenticatedUser user( Auth4cDataXML::Factory::AuthenticatedUser::fromXML( pnodeUser ) );

//		bool bRtn( false );

		try
		{
			/*bRtn =*/ Auth4cDataDB::Factory::passwd( user, authTokenCurrent, authTokenNew );

			std::ostringstream os;
			os << "User " << user.domain().domain() << "/" << user.userId() << " successfully logged on from " << _strHost << ", using profile " << profile.profile();
			Base4c::EventLogger::theLogger()->logMessage( os.str(), Base4c::EventLogger::daemon );
		}
		catch( const Auth4cData::Exception& e )
		{
			std::ostringstream os;
			os << "User " << user.domain().domain() << "/" << user.userId() << " failed to log on from " << _strHost << ", using profile " << profile.profile() << " (" << e.legend() << ")";
			Base4c::EventLogger::theLogger()->logMessage( os.str(), Base4c::EventLogger::daemon );

			throw;
		}

		pnodeResponse->attribute( "response", "passwd" );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::persistSession( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		const Rfc::XMLNode* pnodeAuthenticatedUser = pnodeRequest->findChild( "authenticateduser" );

		if( NULL == pnodeAuthenticatedUser )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No authenticateduser node in persistsession request." );
		}

		Auth4cData::AuthenticatedUser user( Auth4cDataXML::Factory::AuthenticatedUser::fromXML( pnodeAuthenticatedUser ) );
		Auth4cData::Factory::persistSession( user );
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::retrieveSession( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		const Rfc::XMLNode* pnodeSession = pnodeRequest->findChild( "session" );
		if( NULL == pnodeSession )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No session node in retrievesession request." );
		}

		Auth4cData::Session session( Auth4cDataXML::Factory::Session::fromXML( pnodeSession ) );

		Auth4cData::AuthenticatedUser& userAuth( Auth4cData::Factory::retrieveSession( session ) );

		pnodeResponse->attribute( "response", "authenticateduser" );

		Auth4cDataXML::Factory::AuthenticatedUser::toXML( pnodeResponse, userAuth );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::killSession( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		const Rfc::XMLNode* pnodeSession = pnodeRequest->findChild( "session" );
		if( NULL == pnodeSession )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No session node in killsession request." );
		}

		Auth4cData::Session session( Auth4cDataXML::Factory::Session::fromXML( pnodeSession ) );

		Auth4cData::Factory::killSession( session );
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::user( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		const Rfc::XMLNode* pnodeUser = pnodeRequest->findChild( "user" );

		if( NULL == pnodeUser )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No user node in user request." );
		}

		bool bRecurse = pnodeRequest->attribute( "recurse" ) == "Y";

		Auth4cData::User userF( Auth4cDataXML::Factory::User::fromXML( pnodeUser ) );

		Auth4cDataDB::Factory::user( userF, bRecurse );

		pnodeResponse->attribute( "response", "user" );

		Auth4cDataXML::Factory::User::toXML( pnodeResponse, userF, true );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::team( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		const Rfc::XMLNode* pnodeTeam = pnodeRequest->findChild( "team" );

		if( NULL == pnodeTeam )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No team node in team request." );
		}

		bool bReload = pnodeRequest->attribute( "reload" ) == "Y";

		Auth4cData::Team teamF( Auth4cDataXML::Factory::Team::fromXML( pnodeTeam ) );

		Auth4cData::Factory::team( teamF, bReload );

		pnodeResponse->attribute( "response", "team" );

		Auth4cDataXML::Factory::Team::toXML( pnodeResponse, teamF, true );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::teams( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		//bool bAuthenticated( false );

		//try
		//{
		//	if( !_user.session().str().length() )
		//	{
		//		throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_USERNOTFOUND, "There is no authenticated user.", true );
		//	}

		//	bAuthenticated = true;
		//}
		//catch( const Auth4cData::Exception& e )
		//{
		//}

		pnodeResponse->attribute( "response", "teams" );

		Auth4cData::TeamMap teamS;
		for( Auth4cData::TeamMap::const_iterator iterTeam = Auth4cData::Factory::teams().begin();
				iterTeam != Auth4cData::Factory::teams().end();
				++iterTeam )
		{
			//if( iterTeam->second.customer().customer() == _user.domain().customer().customer() )
			//{
			//	teamS[iterTeam->first] = iterTeam->second;
			//}
			//else if( bAuthenticated )
			//{
				teamS[iterTeam->first] = iterTeam->second;
			//}
		}

		Auth4cDataXML::Factory::TeamMap::toXML( pnodeResponse, teamS );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::domains( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		pnodeResponse->attribute( "response", "domains" );

		Auth4cDataXML::Factory::DomainMap::toXML( pnodeResponse, Auth4cData::Factory::domains() );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::profiles( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		pnodeResponse->attribute( "response", "profiles" );

		Auth4cDataXML::Factory::ProfileMap::toXML( pnodeResponse, Auth4cDataDB::Factory::profiles( _strHost ) );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::product( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		const Rfc::XMLNode* pnodeProduct = pnodeRequest->findChild( "product" );
		if( ( NULL == pnodeProduct ) ||
			( pnodeProduct->value().isNull() ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No product node in product request." );
		}

		int nProduct = atoi( pnodeProduct->value() );

		if( !nProduct )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_MALFORMEDXML, "No valid product node in product request." );
		}

		Auth4cData::ProductMap::const_iterator iter = Auth4cData::Factory::products().find( nProduct );
		if( Auth4cData::Factory::products().end() == iter )
		{
			std::ostringstream os;
			os << nProduct;
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_NOSUCHPRODUCT, os.str(), true );
		}

		pnodeResponse->attribute( "response", "product" );

		Auth4cDataXML::Factory::Product::toXML( pnodeResponse, iter->second );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::products( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		if( !_user.session().str().length() )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_USERNOTFOUND, "There is no authenticated user.", true );
		}

		pnodeResponse->attribute( "response", "products" );

		Auth4cDataXML::Factory::ProductMap::toXML( pnodeResponse, Auth4cData::Factory::products() );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::customers( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		pnodeResponse->attribute( "response", "customers" );

		Auth4cDataXML::Factory::CustomerMap::toXML( pnodeResponse, Auth4cData::Factory::customers() );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::who( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest )
{
	try
	{
		if( !_user.session().str().length() )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_USERNOTFOUND, "There is no authenticated user.", true );
		}

		Auth4cData::Permission permissionAdminister_Users( g_product, A4CAuth4C::nA4CP_Administer_Users );

		if( !_user.hasPermission( permissionAdminister_Users ) )
		{
			throw Auth4cData::Exception( __FILE__, __LINE__, Auth4cData::nNAX_PERMISSIONDENIED, "Administer_Users", true );
		}

		pnodeResponse->attribute( "response", "who" );
		
		Auth4cData::AuthenticatedUserMap userS;
		ClientSocket* psocket( NULL );

		try
		{
			theApp.sockets().lock();
			psocket = theApp.sockets().next( psocket );
			while( NULL != psocket )
			{
				if( ClientSocket::clientConnect == psocket->type() )
				{
					try
					{
						userS[psocket->user().key()] = psocket->user();
					}
					catch( ... )
					{
						// The user isn't yet authenticated
					}
				}

				psocket = theApp.sockets().next( psocket );
			}

			theApp.sockets().unlock();
		}
		catch( ... )
		{
			theApp.sockets().unlock();
			throw;
		}

		Auth4cDataXML::Factory::AuthenticatedUserMap::toXML( pnodeResponse, userS );

#if defined (DEBUG)
		std::cout << (const char*)pnodeResponse->stream() << "\n";
#endif
	}
	catch( const Auth4cData::Exception& e )
	{
		std::cout << e << std::endl;
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
	}
	catch( const Base4c::Exception& e )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, e.legend() );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
	catch( ... )
	{
		Auth4cData::Exception e1( __FILE__, __LINE__, Auth4cData::nNAX_ALIENEXCEPTION, "..." );
		pnodeResponse->attribute( "response", "exception" );
		Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e1 );
	}
}

void ClientSocket::unknownRequest( Rfc::XMLNode* pnodeResponse, const Rfc::XMLNode* pnodeRequest, const Rfc::String& strRequest )
{
	Auth4cData::Exception e( __FILE__, __LINE__, Auth4cData::nNAX_BADREQUEST, (const char*)strRequest, true );
	pnodeResponse->attribute( "response", "exception" );
	Auth4cDataXML::Factory::Exception::toXML( pnodeResponse, e );
}
